import logging

from eth_utils import add_0x_prefix
from contracts_lib_py.utils import process_tx_receipt
from contracts_lib_py.web3_provider import Web3Provider
from common_utils_py.did import did_to_id
from common_utils_py.did_resolver.did_resolver import DIDResolver

from nevermined_sdk_py.gateway import GatewayProvider
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from nevermined_sdk_py.secret_store import SecretStoreProvider

logger = logging.getLogger(__name__)


def refund_reward(event, agreement_id, did, service_agreement, price, consumer_account,
                  publisher_address, condition_ids, escrow_condition_id):
    """
    Refund the reward to the publisher address.

    :param event: AttributeDict with the event data.
    :param agreement_id: id of the agreement, hex str
    :param did: DID, str
    :param service_agreement: ServiceAgreement instance
    :param price: Asset price, int
    :param consumer_account: Account instance of the consumer
    :param publisher_address: ethereum account address of publisher, hex str
    :param condition_ids: is a list of bytes32 content-addressed Condition IDs, bytes32
    :param escrow_condition_id: hex str the id of escrow reward condition at this `agreement_id`
    """
    logger.debug(f"trigger refund (agreement {agreement_id}) after event {event}.")
    if Keeper.get_instance().condition_manager.get_condition_state(escrow_condition_id) > 1:
        logger.debug(
            f'escrow payment condition already fulfilled/aborted: '
            f'agreementId={agreement_id}, escrow reward conditionId={escrow_condition_id},'
            f' publisher={publisher_address}'
        )
        return

    access_id, lock_id = condition_ids[:2]
    name_to_parameter = {param.name: param for param in
                         service_agreement.condition_by_name['escrowPayment'].parameters}
    document_id = add_0x_prefix(name_to_parameter['_documentId'].value)
    asset_id = add_0x_prefix(did_to_id(did))
    did_owner = Keeper.get_instance().agreement_manager.get_agreement_did_owner(agreement_id)
    assert document_id == asset_id, f'document_id {document_id} <=> asset_id {asset_id} mismatch.'
    assert price == service_agreement.get_price(), 'price mismatch.'
    try:
        escrow_condition = Keeper.get_instance().escrow_payment_condition
        tx_hash = escrow_condition.fulfill(
            agreement_id,
            price,
            Web3Provider.get_web3().toChecksumAddress(did_owner),
            consumer_account.address,
            lock_id,
            access_id,
            consumer_account
        )
        process_tx_receipt(
            tx_hash,
            getattr(escrow_condition.contract.events, escrow_condition.FULFILLED_EVENT)(),
            'EscrowReward.Fulfilled'
        )
    except Exception as e:
        logger.error(
            f'Error when doing escrow_payment_condition.fulfills (agreementId {agreement_id}): {e}',
            exc_info=1)
        raise e


def consume_asset(event, agreement_id, did, service_agreement, consumer_account, consume_callback,
                  secret_store_url, parity_url, downloads_path):
    """
    Consumption of an asset after get the event call.

    :param event: AttributeDict with the event data.
    :param agreement_id: id of the agreement, hex str
    :param did: DID, str
    :param service_agreement: ServiceAgreement instance
    :param consumer_account: Account instance of the consumer
    :param consume_callback:
    :param secret_store_url: str URL of secret store node for retrieving decryption keys
    :param parity_url: str URL of parity client to use for secret store encrypt/decrypt
    :param downloads_path: str path to save downloaded files
    """
    logger.debug(f"consuming asset (agreementId {agreement_id}) after event {event}.")
    if consume_callback:
        secret_store = SecretStoreProvider.get_secret_store(
            secret_store_url, parity_url, consumer_account
        )
        gateway = GatewayProvider.get_gateway()

        consume_callback(
            agreement_id,
            service_agreement.service_definition_id,
            DIDResolver(Keeper.get_instance().did_registry).resolve(did),
            consumer_account,
            downloads_path,
            gateway,
            secret_store
        )



import os

from contracts_lib_py.web3_provider import Web3Provider
from web3 import Web3
from contracts_lib_py.web3.http_provider import CustomHTTPProvider
import pytest

from nevermined_sdk_py.config_provider import ConfigProvider
from nevermined_sdk_py.config import Config
from nevermined_sdk_py.nevermined.nevermined import Nevermined

INFURA_TOKEN = os.environ.get("INFURA_TOKEN")


@pytest.mark.parametrize("keeper_url,network_name", [
    [f"https://mainnet.infura.io/v3/{INFURA_TOKEN}", ["mainnet"]],
    [f"https://rinkeby.infura.io/v3/{INFURA_TOKEN}", ["rinkeby"]],
    ("http://localhost:8545", ["spree", "geth-localnet"]),
    (f"https://polygon-mumbai.infura.io/v3/{INFURA_TOKEN}", ["mumbai"]),
    ("https://alfajores-forno.celo-testnet.org", ["celo-alfajores"])
])
def test_artifact(keeper_url, network_name):
    options = {
        "keeper-contracts": {
            "keeper.url": keeper_url
        }
    }
    config = Config(options_dict=options)
    Web3Provider._web3 = Web3(CustomHTTPProvider(config.keeper_url))
    ConfigProvider.set_config(config)

    nevermined = Nevermined()
    assert nevermined.keeper.network_name in network_name


import os

import pytest
from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceAuthorizationTypes, ServiceTypes
from common_utils_py.ddo.ddo import DDO
from contracts_lib_py.web3_provider import Web3Provider

from examples import ExampleConfig
from nevermined_sdk_py import ConfigProvider
from nevermined_sdk_py.nevermined.agreements import check_token_address
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from tests.resources.helper_functions import (get_consumer_account, get_publisher_account,
                                              get_registered_ddo, get_registered_with_psk, log_event)


def test_buy_asset(publisher_instance_no_init, consumer_instance_no_init):
    config = ExampleConfig.get_config()
    keeper = Keeper.get_instance()
    # :TODO: enable the actual SecretStore
    # SecretStoreProvider.set_secret_store_class(SecretStore)
    pub_acc = get_publisher_account()

    # Register ddo
    ddo = get_registered_ddo(publisher_instance_no_init, pub_acc)
    assert isinstance(ddo, DDO)
    # nevermined here will be used only to publish the asset. Handling the asset by the publisher
    # will be performed by the Gateway server running locally

    # restore the http client because we want the actual Gateway server to do the work
    # not the GatewayMock.
    # Gateway.set_http_client(requests)
    consumer_account = get_consumer_account()

    downloads_path_elements = len(
        os.listdir(consumer_instance_no_init._config.downloads_path)) if os.path.exists(
        consumer_instance_no_init._config.downloads_path) else 0
    # sign agreement using the registered asset did above
    service = ddo.get_service(service_type=ServiceTypes.ASSET_ACCESS)
    sa = ServiceAgreement.from_service_dict(service.as_dictionary())
    # This will send the access request to Gateway which in turn will execute the agreement on-chain
    consumer_instance_no_init.accounts.request_tokens(consumer_account, 100)

    agreement_id = consumer_instance_no_init.assets.order(
        ddo.did, sa.index, consumer_account, consumer_account)

    event_wait_time = 10
    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        event_wait_time,
        log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for LockRewardCondition.Fulfilled'

    # give access
    publisher_instance_no_init.agreements.conditions.grant_access(
        agreement_id, ddo.asset_id, consumer_account.address, pub_acc)
    event = keeper.access_condition.subscribe_condition_fulfilled(
        agreement_id,
        event_wait_time,
        log_event(keeper.access_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessCondition.Fulfilled'
    assert consumer_instance_no_init.agreements.is_access_granted(agreement_id, ddo.did, consumer_account.address)

    amounts = list(map(int, service.get_param_value_by_name('_amounts')))
    receivers = service.get_param_value_by_name('_receivers')
    token_address = check_token_address(
        keeper, service.get_param_value_by_name('_tokenAddress'))

    publisher_instance_no_init.agreements.conditions.release_reward(
        agreement_id, ddo.asset_id, amounts, receivers, consumer_account.address, token_address, pub_acc)

    assert consumer_instance_no_init.assets.access(
        agreement_id,
        ddo.did,
        sa.index,
        consumer_account,
        config.downloads_path)

    assert len(os.listdir(config.downloads_path)) == downloads_path_elements + 1

    # Check that we can access only an specific file in passing the index.
    assert consumer_instance_no_init.assets.access(
        agreement_id,
        ddo.did,
        sa.index,
        consumer_account,
        config.downloads_path,
        1
    )
    assert len(os.listdir(config.downloads_path)) == downloads_path_elements + 1

    with pytest.raises(AssertionError):
        consumer_instance_no_init.assets.access(
            agreement_id,
            ddo.did,
            sa.index,
            consumer_account,
            config.downloads_path,
            -2
        )

    with pytest.raises(AssertionError):
        consumer_instance_no_init.assets.access(
            agreement_id,
            ddo.did,
            sa.index,
            consumer_account,
            config.downloads_path,
            3
        )

    # decrypt the contentUrls using the publisher account instead of consumer account.
    # if the secret store is working and ACL check is enabled, this should fail
    # since SecretStore decrypt will fail the checkPermissions check and the gateway will return
    # an http error code back
    with pytest.raises(ValueError) as e:
        consumer_instance_no_init.assets.access(
            agreement_id,
            ddo.did,
            sa.index,
            pub_acc,
            config.downloads_path
        )
        print(e)


@pytest.mark.skip(reason="Run in local with the gateway up")
def test_buy_asset_no_secret_store(publisher_instance_gateway, consumer_instance_gateway):
    config = ExampleConfig.get_config()
    ConfigProvider.set_config(config)
    keeper = Keeper.get_instance()

    w3 = Web3Provider.get_web3()
    pub_acc = get_publisher_account()

    for method in [ServiceAuthorizationTypes.PSK_ECDSA, ServiceAuthorizationTypes.PSK_RSA]:
        # Register ddo
        ddo = get_registered_with_psk(publisher_instance_gateway, pub_acc, auth_method=method)
        assert isinstance(ddo, DDO)
        # nevermined here will be used only to publish the asset. Handling the asset by the publisher
        # will be performed by the Gateway server running locally

        # restore the http client because we want the actual Gateway server to do the work
        # not the GatewayMock.
        # Gateway.set_http_client(requests)
        consumer_account = get_consumer_account()

        downloads_path_elements = len(
            os.listdir(consumer_instance_gateway._config.downloads_path)) if os.path.exists(
            consumer_instance_gateway._config.downloads_path) else 0
        # sign agreement using the registered asset did above
        service = ddo.get_service(service_type=ServiceTypes.ASSET_ACCESS)
        sa = ServiceAgreement.from_service_dict(service.as_dictionary())
        # This will send the access request to Gateway which in turn will execute the agreement on-chain
        consumer_instance_gateway.accounts.request_tokens(consumer_account, 100)
        agreement_id = consumer_instance_gateway.assets.order(
            ddo.did, sa.index, consumer_account, consumer_account)

        event_wait_time = 10
        event = keeper.access_template.subscribe_agreement_created(
            agreement_id,
            event_wait_time,
            log_event(keeper.access_template.AGREEMENT_CREATED_EVENT),
            (),
            wait=True
        )
        assert event, 'no event for EscrowAccessSecretStoreTemplate.AgreementCreated'

        event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
            agreement_id,
            event_wait_time,
            log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
            (),
            wait=True
        )
        assert event, 'no event for LockRewardCondition.Fulfilled'

        # give access
        publisher_instance_gateway.agreements.conditions.grant_access(
            agreement_id, ddo.asset_id, consumer_account.address, pub_acc)
        event = keeper.access_condition.subscribe_condition_fulfilled(
            agreement_id,
            event_wait_time,
            log_event(keeper.access_condition.FULFILLED_EVENT),
            (),
            wait=True
        )
        assert event, 'no event for AccessSecretStoreCondition.Fulfilled'
        assert consumer_instance_gateway.agreements.is_access_granted(agreement_id, ddo.did, consumer_account.address)

        token_address = check_token_address(
            keeper, sa.get_param_value_by_name('_tokenAddress'))

        publisher_instance_gateway.agreements.conditions.release_reward(
            agreement_id, ddo.asset_id, sa.get_amounts_int(), sa.get_receivers(), token_address, pub_acc)

        assert consumer_instance_gateway.assets.access(
            agreement_id,
            ddo.did,
            sa.index,
            consumer_account,
            config.downloads_path)

        assert len(os.listdir(config.downloads_path)) == downloads_path_elements + 1

        with pytest.raises(AssertionError):
            consumer_instance_gateway.assets.access(
                agreement_id,
                ddo.did,
                sa.index,
                consumer_account,
                config.downloads_path,
                -2
            )

        with pytest.raises(AssertionError):
            consumer_instance_gateway.assets.access(
                agreement_id,
                ddo.did,
                sa.index,
                consumer_account,
                config.downloads_path,
                3
            )



import logging
import logging
import os

from common_utils_py.did import did_to_id, convert_to_bytes

logger = logging.getLogger(__name__)


class NFTs:
    """Nevermined NFTs class."""

    def __init__(self, keeper):
        self._keeper = keeper

    def mint(self, did, amount, account):
        """
        Mint an amount of nfts.
        :param did: the id of an asset on-chain, hex str
        :param amount: amount of nft to be minted, int
        :param account: Account executing the action
        """
        return self._keeper.did_registry.mint(convert_to_bytes(did), amount, account)

    def burn(self, did, amount, account):
        """
        Burn an amount of nfts.
        :param did: the id of an asset on-chain, hex str
        :param amount: amount of nft to be burnt, int
        :param account: Account executing the action
        """
        return self._keeper.did_registry.burn(convert_to_bytes(did), amount, account)

    def transfer_nft(self, did, address, amount, account):
        """
        Transfer nft to another address. Return true if successful

        :param did: the id of an asset on-chain, hex str
        :param address: ethereum account address, hex str
        :param amount: amount of nft to be transfer, int
        :param account: Account executing the action

        """
        return self._keeper.nft_upgradeable.transfer_nft(did_to_id(did), address, amount, account)

    def balance(self, address, did):
        """
        Return nft balance.

        :param address: ethereum account address, hex str
        :param did: the id of an asset on-chain, hex str
        """
        return self._keeper.nft_upgradeable.balance(address, did_to_id(did))





import logging
import time

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes, ServiceTypesIndices
from common_utils_py.did import did_to_id
from common_utils_py.exceptions import (
    InvalidAgreementTemplate,
    ServiceAgreementExists,
)
from contracts_lib_py.utils import add_ethereum_prefix_and_hash_msg
from contracts_lib_py.web3_provider import Web3Provider

from nevermined_sdk_py.agreement_events.access_agreement import consume_asset, refund_reward
from nevermined_sdk_py.agreement_events.compute_agreement import execute_computation
from nevermined_sdk_py.agreement_events.payments import \
    fulfillLockRewardCondition
from nevermined_sdk_py.nevermined.conditions import Conditions
from web3 import Web3

logger = logging.getLogger(__name__)

ZERO_ADDRESS = '0x0000000000000000000000000000000000000000'


def check_token_address(keeper, token_address):
    if token_address == '0x0' or token_address == ZERO_ADDRESS:
        return ZERO_ADDRESS
    elif not Web3.isAddress(token_address):
        return keeper.token.address
    return token_address


class Agreements:
    """Nevermined agreements class."""

    def __init__(self, keeper, asset_resolver, asset_consumer, asset_executor, config):
        self._keeper = keeper
        self._asset_resolver = asset_resolver
        self._asset_consumer = asset_consumer
        self._asset_executor = asset_executor
        self._config = config
        self.conditions = Conditions(self._keeper)

    def prepare(self, did, consumer_account, service_index):
        """

        :param did: str representation fo the asset DID. Use this to retrieve the asset DDO.
        :param consumer_account: Account instance of the consumer
        :param service_index: int identifies the specific service in
         the ddo to use in this agreement.
        :return: tuple (agreement_id: str, signature: hex str)
        """
        agreement_id_seed = ServiceAgreement.create_new_agreement_id()
        agreement_id = self._keeper.agreement_manager.hash_id(agreement_id_seed, consumer_account.address)
        signature = self._sign(agreement_id, did, consumer_account, service_index)
        return (agreement_id_seed, agreement_id), signature

    def create(self, did, index, agreement_id_seed, consumer_acc, account):
        """
        Execute the service agreement on-chain using keeper's ServiceAgreement contract.

        The on-chain executeAgreement method requires the following arguments:
        templateId, signature, consumer, hashes, timeouts, serviceAgreementId, did.
        `agreement_message_hash` is necessary to verify the signature.
        The consumer `signature` includes the conditions timeouts and parameters values which
        is usedon-chain to verify that the values actually match the signed hashes.

        :param did: str representation fo the asset DID. Use this to retrieve the asset DDO.
        :param index: str identifies the specific service in
         the ddo to use in this agreement.
        :param agreement_id_seed: 32 bytes identifier created by the consumer and will be used
         on-chain for the executed agreement.
         conditions and their parameters values and other details of the agreement.
        :param consumer_address: ethereum account address of consumer, hex str
        :param account: Account instance creating the agreement. Can be either the
            consumer, publisher or provider
        :return: dict the `executeAgreement` transaction receipt
        """
        consumer_address = consumer_acc.address
        consumer_address_real = consumer_acc.address
        babyjub_address = None
        assert consumer_address and Web3Provider.get_web3().isChecksumAddress(
            consumer_address), f'Invalid consumer address {consumer_address}'

        payment_involved = True
        asset = self._asset_resolver.resolve(did)
        asset_id = asset.asset_id
        service = asset.get_service_by_index(index)
        if service.type == ServiceTypes.ASSET_ACCESS:
            agreement_template = self._keeper.access_template
            template_address = self._keeper.access_template.address
        elif service.type == ServiceTypes.ASSET_ACCESS_PROOF:
            agreement_template = self._keeper.access_proof_template
            template_address = self._keeper.access_proof_template.address
            babyjub_address = consumer_acc.babyjub_address
        elif service.type == ServiceTypes.CLOUD_COMPUTE:
            agreement_template = self._keeper.escrow_compute_execution_template
            template_address = self._keeper.escrow_compute_execution_template.address
        elif service.type == ServiceTypes.NFT_SALES:
            agreement_template = self._keeper.nft_sales_template
            template_address = self._keeper.nft_sales_template.address
        elif service.type == ServiceTypes.NFT_ACCESS:
            payment_involved = False
            agreement_template = self._keeper.nft_access_template
            template_address = self._keeper.nft_access_template.address
        else:
            raise Exception('The agreement could not be created. Review the index of your service.')

        agreement_template_approved = self._keeper.template_manager.is_template_approved(template_address)
        if not agreement_template_approved:
            msg = (f'The Service Agreement Template contract at address '
                   f'{template_address} is not '
                   f'approved and cannot be used for creating service agreements.')
            logger.warning(msg)
            raise InvalidAgreementTemplate(msg)

        service_agreement = ServiceAgreement.from_service_index(index, asset)
        token_address = check_token_address(
            self._keeper, service_agreement.get_param_value_by_name('_tokenAddress'))

        publisher_address = Web3Provider.get_web3().toChecksumAddress(asset.publisher)

        ((agreement_id_seed, agreement_id), *conditions) = service_agreement.generate_agreement_condition_ids(
            agreement_id_seed, asset_id, consumer_address, self._keeper, token_address=token_address,
            babyjub_pk=babyjub_address)

        if agreement_template.get_agreement_consumer(agreement_id) != ZERO_ADDRESS:
            raise ServiceAgreementExists(
                f'Service agreement {agreement_id} already exists, cannot reuse '
                f'the same agreement id.')

        condition_ids = [c[0] for c in conditions]
        time_locks = service_agreement.conditions_timelocks
        time_outs = service_agreement.conditions_timeouts

        if payment_involved:
            price = service_agreement.get_price()
            consumer_balance = self._keeper.token.get_token_balance(consumer_address_real)
            if price > consumer_balance:
                return Exception(
                    f'The consumer balance is '
                    f'{consumer_balance}. '
                    f'This balance is lower that the asset price {price}.')

        if service.type == ServiceTypes.NFT_SALES:
            conditions_ordered = [condition_ids[1], condition_ids[0], condition_ids[2]]
        elif service.type == ServiceTypes.NFT_ACCESS:
            conditions_ordered = [condition_ids[1], condition_ids[0]]
        else:
            conditions_ordered = condition_ids

        success = agreement_template.create_agreement(
            agreement_id_seed,
            asset_id,
            conditions_ordered,
            time_locks,
            time_outs,
            Web3.toChecksumAddress(account.address),
            account
        )

        if success:
            if payment_involved:
                self.conditions.lock_payment(
                    agreement_id,
                    asset_id,
                    service_agreement.get_amounts_int(),
                    service_agreement.get_receivers(),
                    token_address,
                    account)
                return self._is_condition_fulfilled(agreement_id, 'lockReward')
            return True
        return False

    def status(self, agreement_id):
        """
        Get the status of a service agreement.

        :param agreement_id: id of the agreement, hex str
        :return: dict with condition status of each of the agreement's conditions or None if the
        agreement is invalid.
        """
        condition_ids = self._keeper.agreement_manager.get_agreement(agreement_id).condition_ids
        result = {"agreementId": agreement_id}
        conditions = dict()
        for i in condition_ids:
            conditions[self._keeper.get_condition_name_by_address(
                self._keeper.condition_manager.get_condition(
                    i).type_ref)] = self._keeper.condition_manager.get_condition_state(i)
        result["conditions"] = conditions
        return result

    def is_access_granted(self, agreement_id, did, consumer_address):
        """
        Check permission for the agreement.

        Verify on-chain that the `consumer_address` has permission to access the given asset `did`
        according to the `agreement_id`.

        :param agreement_id: id of the agreement, hex str
        :param did: DID, str
        :param consumer_address: ethereum account address of consumer, hex str
        :return: bool True if user has permission
        """
        agreement_consumer = self._keeper.access_template.get_agreement_consumer(
            agreement_id)

        if agreement_consumer is None:
            return False

        if agreement_consumer != consumer_address:
            logger.warning(f'Invalid consumer address {consumer_address} and/or '
                           f'service agreement id {agreement_id} (did {did})'
                           f', agreement consumer is {agreement_consumer}')
            return False

        document_id = did_to_id(did)
        return self._keeper.access_condition.check_permissions(
            document_id, consumer_address
        )

    def _sign(self, agreement_id, did, consumer_account, service_index):
        """
        Sign a service agreement.

        :param agreement_id: 32 bytes identifier created by the consumer and will be used
         on-chain for the executed agreement.
        :param did: str representation fo the asset DID. Use this to retrieve the asset DDO.
        :param consumer_account: Account instance of the consumer
        :param service_index: int identifies the specific service in
         the ddo to use in this agreement.
        :return: signature
        """
        asset = self._asset_resolver.resolve(did)
        service_agreement = asset.get_service_by_index(service_index)

        publisher_address = self._keeper.did_registry.get_did_owner(asset.asset_id)
        if service_index == ServiceTypesIndices.DEFAULT_ACCESS_PROOF_INDEX:
            agreement_hash = service_agreement.get_service_agreement_hash(
                agreement_id, asset.asset_id, consumer_account.address, publisher_address, self._keeper,
                babyjub_pk=consumer_account.babyjub_address
            )
        else:
            agreement_hash = service_agreement.get_service_agreement_hash(
                agreement_id, asset.asset_id, consumer_account.address, publisher_address, self._keeper
            )
        signature = self._keeper.sign_hash(add_ethereum_prefix_and_hash_msg(agreement_hash),
                                           consumer_account)
        address = self._keeper.personal_ec_recover(agreement_hash, signature)
        assert address == consumer_account.address
        logger.debug(f'agreement-signature={signature}, agreement-hash={agreement_hash}')
        return signature

    def _is_condition_fulfilled(self, agreement_id, condition_type):
        status = self.status(agreement_id)
        condition_status = status.get('conditions').get(condition_type)
        logger.debug(f'Condition check[  ${condition_type}  ] : + ${condition_status}')

        return condition_status == 2

    def _process_consumer_agreement_events(
            self, agreement_id, did, service_agreement, account,
            condition_ids, publisher_address, from_block, auto_consume, agreement_type):
        logger.debug(
            f'process consumer events for agreement {agreement_id}, blockNumber {from_block + 10}')
        if agreement_type == ServiceTypes.ASSET_ACCESS:
            self._keeper.access_template.subscribe_agreement_created(
                agreement_id,
                300,
                fulfillLockRewardCondition,
                (agreement_id, service_agreement.get_price(), account, condition_ids[1]),
                from_block=from_block
            )
        else:
            self._keeper.escrow_compute_execution_template.subscribe_agreement_created(
                agreement_id,
                300,
                fulfillLockRewardCondition,
                (agreement_id, service_agreement.get_price(), account, condition_ids[1]),
                from_block=from_block
            )

        if auto_consume:
            def _refund_callback(_price, _publisher_address, _condition_ids):
                def do_refund(_event, _agreement_id, _did, _service_agreement, _consumer_account,
                              *_):
                    refund_reward(
                        _event, _agreement_id, _did, _service_agreement, _price,
                        _consumer_account, _publisher_address, _condition_ids, _condition_ids[2]
                    )

                return do_refund

            conditions_dict = service_agreement.condition_by_name
            if agreement_type == ServiceTypes.ASSET_ACCESS:
                self._keeper.access_condition.subscribe_condition_fulfilled(
                    agreement_id,
                    max(conditions_dict['accessSecretStore'].timeout, 300),
                    consume_asset,
                    (agreement_id, did, service_agreement, account, self._asset_consumer.download,
                     self._config.secret_store_url, self._config.parity_url,
                     self._config.downloads_path),
                    timeout_callback=_refund_callback(
                        service_agreement.get_price(), publisher_address, condition_ids
                    ),
                    from_block=from_block
                )
            else:
                self._keeper.compute_execution_condition.subscribe_condition_fulfilled(
                    agreement_id,
                    max(conditions_dict['execCompute'].timeout, 300),
                    execute_computation,
                    (agreement_id, did, service_agreement, account, self._asset_executor.execute),
                    timeout_callback=_refund_callback(
                        service_agreement.get_price(), publisher_address, condition_ids
                    ),
                    from_block=from_block
                )

    def _log_agreement_info(self, asset, service_agreement, agreement_id, agreement_signature,
                            consumer_address, publisher_account, condition_ids):
        agreement_hash = service_agreement.get_service_agreement_hash(
            agreement_id, asset.asset_id, consumer_address, publisher_account.address, self._keeper)
        publisher_ether_balance = self._keeper.get_ether_balance(publisher_account.address)
        logger.debug(
            f'Agreement parameters:'
            f'\n  agreement id: {agreement_id}'
            f'\n  consumer address: {consumer_address}'
            f'\n  publisher address: {publisher_account.address}'
            f'\n  conditions ids: {condition_ids}'
            f'\n  asset did: {asset.did}'
            f'\n  agreement signature: {agreement_signature}'
            f'\n  agreement hash: {agreement_hash}'
            f'\n  EscrowAccessSecretStoreTemplate: '
            f'{self._keeper.access_template.address}'
            f'\n  publisher ether balance: {publisher_ether_balance}'
        )

    def _verify_service_agreement_signature(self, did, agreement_id, service_index,
                                            consumer_address, signature, ddo=None):
        """
        Verify service agreement signature.

        Verify that the given signature is truly signed by the `consumer_address`
        and represents this did's service agreement..

        :param did: DID, str
        :param agreement_id: id of the agreement, hex str
        :param service_index: identifier of the service inside the asset DDO, str
        :param consumer_address: ethereum account address of consumer, hex str
        :param signature: Signature, str
        :param ddo: DDO instance
        :return: True if signature is legitimate, False otherwise
        :raises: ValueError if service is not found in the ddo
        :raises: AssertionError if conditions keys do not match the on-chain conditions keys
        """
        if not ddo:
            ddo = self._asset_resolver.resolve(did)

        service_agreement = ddo.get_service_by_index(service_index)
        agreement_hash = service_agreement.get_service_agreement_hash(
            agreement_id, ddo.asset_id, consumer_address,
            Web3Provider.get_web3().toChecksumAddress(ddo.proof['creator']), self._keeper)

        recovered_address = self._keeper.personal_ec_recover(agreement_hash, signature)
        is_valid = (recovered_address == consumer_address)
        if not is_valid:
            logger.warning(f'Agreement signature failed: agreement hash is {agreement_hash.hex()}')

        return is_valid

    def _approve_token_transfer(self, amount, consumer_account):
        if self._keeper.token.get_token_balance(consumer_account.address) < amount:
            raise ValueError(
                f'Account {consumer_account.address} does not have sufficient tokens '
                f'to approve for transfer.')

        self._keeper.token.token_approve(self._keeper.payment_conditions.address, amount,
                                         consumer_account)

    def _get_agreement(self, agreement_id):
        """
        Retrieve the agreement data of agreement_id.

        :param agreement_id: id of the agreement, hex str
        :return: AgreementValues instance -- a namedtuple with the following attributes:

            did,
            owner,
            template_id,
            condition_ids,
            updated_by,
            block_number_updated

        """
        return self._keeper.agreement_manager.get_agreement(agreement_id)



import logging

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes

logger = logging.getLogger(__name__)


class AssetExecutor:
    """Class representing the call to the Gateway execute endpoint."""

    @staticmethod
    def execute(agreement_id, compute_ddo, workflow_ddo, consumer_account, gateway, index, config):
        """

        :param agreement_id:
        :param workflow_ddo:
        :param consumer_account:
        :param index:
        :return: the id of the compute job
        """
        service_endpoint = ServiceAgreement.from_ddo(ServiceTypes.CLOUD_COMPUTE,
                                                     compute_ddo).service_endpoint
        response = gateway.execute_compute_service(agreement_id, service_endpoint,
                                                   consumer_account, workflow_ddo, config)

        return response.json()["workflowId"]



import copy
import json
import logging
import os

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_factory import ServiceDescriptor, ServiceFactory
from common_utils_py.agreements.service_types import ServiceAuthorizationTypes, ServiceTypes
from common_utils_py.ddo.ddo import DDO
from common_utils_py.ddo.metadata import MetadataMain
from common_utils_py.ddo.public_key_rsa import PUBLIC_KEY_TYPE_RSA
from common_utils_py.did import DID, did_to_id, did_to_id_bytes, convert_to_bytes
from common_utils_py.exceptions import (
    DIDAlreadyExist,
)
from common_utils_py.metadata.exceptions import MetadataGenericError
from common_utils_py.metadata.metadata_provider import MetadataProvider
from common_utils_py.utils.utilities import checksum
from contracts_lib_py.utils import add_ethereum_prefix_and_hash_msg
from contracts_lib_py.web3_provider import Web3Provider
from eth_utils import add_0x_prefix

from nevermined_sdk_py.gateway.gateway_provider import GatewayProvider
from nevermined_sdk_py.secret_store.secret_store_provider import SecretStoreProvider

logger = logging.getLogger(__name__)


class Assets:
    """Nevermined assets class."""

    def __init__(self, keeper, did_resolver, agreements, asset_consumer, asset_executor, config):
        self._keeper = keeper
        self._did_resolver = did_resolver
        self._agreements = agreements
        self._asset_consumer = asset_consumer
        self._asset_executor = asset_executor
        self._config = config
        self._metadata_url = config.metadata_url

        downloads_path = os.path.join(os.getcwd(), 'downloads')
        if self._config.has_option('resources', 'downloads.path'):
            downloads_path = self._config.get('resources', 'downloads.path') or downloads_path
        self._downloads_path = downloads_path

    def _get_metadata_provider(self, url=None, account=None):
        return MetadataProvider.get_metadata_provider(url or self._metadata_url, account)

    def _get_secret_store(self, account):
        return SecretStoreProvider.get_secret_store(
            self._config.secret_store_url, self._config.parity_url, account
        )

    def create(self, metadata, publisher_account,
               service_descriptors=None, providers=None,
               authorization_type=ServiceAuthorizationTypes.PSK_RSA, use_secret_store=False,
               activity_id=None, attributes=None, asset_rewards={"_amounts": [], "_receivers": [], "_tokenAddress": ""},
               cap=None, royalties=None, mint=0, timeout=3600):
        """
        Register an asset in both the keeper's DIDRegistry (on-chain) and in the Metadata store.

        :param metadata: dict conforming to the Metadata accepted by Nevermined Protocol.
        :param publisher_account: Account of the publisher registering this asset
        :param service_descriptors: list of ServiceDescriptor tuples of length 2.
            The first item must be one of ServiceTypes and the second
            item is a dict of parameters and values required by the service
        :param providers: list of addresses of providers of this asset (a provider is
            an ethereum account that is authorized to provide asset services)
        :param authorization_type: str indicate the authorization type that is going to be used
        to encrypt the urls.
            SecretStore, PSK-RSA and PSK-ECDSA are supported.
        :param use_secret_store: bool indicate whether to use the secret store directly for
            encrypting urls (Uses Gateway provider service if set to False)
        :param activity_id: identifier of the activity creating the new entity
        :param attributes: attributes associated with the action
        :param asset_rewards: rewards distribution including the amounts and the receivers
        :param cap: max cap of nfts that can be minted for the asset
        :param royalties: royalties in the secondary market going to the original creator
        :param mint: number of nfts to mint just after registration of the asset
        :return: DDO instance
        """
        assert isinstance(metadata, dict), f'Expected metadata of type dict, got {type(metadata)}'

        # copy metadata so we don't change the original
        metadata_copy = copy.deepcopy(metadata)

        if '_tokenAddress' not in asset_rewards:
            asset_rewards['_tokenAddress'] = ''

        # Create a DDO object
        ddo = DDO()
        gateway = GatewayProvider.get_gateway()
        ddo_service_endpoint = self._get_metadata_provider(account=publisher_account).get_service_endpoint()
        metadata_service_desc = ServiceDescriptor.metadata_service_descriptor(metadata_copy,
                                                                              ddo_service_endpoint)
        if metadata_copy['main']['type'] == 'dataset' or metadata_copy['main'][
            'type'] == 'algorithm':
            access_service_attributes = self._build_access(metadata_copy, publisher_account, asset_rewards, timeout)
            if not service_descriptors:
                if authorization_type == ServiceAuthorizationTypes.PSK_RSA:
                    service_descriptors = [ServiceDescriptor.authorization_service_descriptor(
                        self._build_authorization(authorization_type,
                                                  public_key=gateway.get_rsa_public_key(self._config)),
                        gateway.get_access_endpoint(self._config)
                    )]
                elif authorization_type == ServiceAuthorizationTypes.PSK_ECDSA:
                    service_descriptors = [ServiceDescriptor.authorization_service_descriptor(
                        self._build_authorization(authorization_type,
                                                  public_key=gateway.get_ecdsa_public_key(self._config)),
                        gateway.get_access_endpoint(self._config)
                    )]
                else:
                    service_descriptors = [ServiceDescriptor.authorization_service_descriptor(
                        self._build_authorization(authorization_type, threshold=0),
                        self._config.secret_store_url
                    )]
                service_descriptors += [ServiceDescriptor.access_service_descriptor(
                    access_service_attributes,
                    gateway.get_access_endpoint(self._config)
                )]
            else:
                service_types = set(map(lambda x: x[0], service_descriptors))
                if ServiceTypes.AUTHORIZATION not in service_types:
                    if authorization_type == ServiceAuthorizationTypes.PSK_RSA:
                        service_descriptors += [ServiceDescriptor.authorization_service_descriptor(
                            self._build_authorization(authorization_type,
                                                      public_key=gateway.get_rsa_public_key(self._config)),
                            gateway.get_access_endpoint(self._config)
                        )]
                    elif authorization_type == ServiceAuthorizationTypes.PSK_ECDSA:
                        service_descriptors += [ServiceDescriptor.authorization_service_descriptor(
                            self._build_authorization(authorization_type,
                                                      public_key=gateway.get_ecdsa_public_key(self._config)),
                            gateway.get_access_endpoint(self._config)
                        )]
                    else:
                        service_descriptors += [ServiceDescriptor.authorization_service_descriptor(
                            self._build_authorization(authorization_type, threshold=0),
                            self._config.secret_store_url
                        )]
                else:
                    service_descriptors += [ServiceDescriptor.access_service_descriptor(
                        access_service_attributes,
                        gateway.get_access_endpoint(self._config)

                    )]
        else:
            if not service_descriptors:
                service_descriptors = []
            else:
                service_descriptors += []
            logger.info('registering a workflow.')
        # Add all services to ddo
        service_descriptors = [metadata_service_desc] + service_descriptors

        services = ServiceFactory.build_services(service_descriptors)
        checksums = dict()
        for service in services:
            checksums[str(service.index)] = checksum(service.main)

        # Adding proof to the ddo.
        ddo.add_proof(checksums, publisher_account)

        # Generating the did and adding to the ddo.
        did_seed = checksum(ddo.proof['checksum'])
        asset_id = self._keeper.did_registry.hash_did(did_seed, publisher_account.address)
        ddo._did = DID.did(asset_id)
        did = ddo._did

        logger.debug(f'Generating new did: {did}')
        # Check if it's already registered first!
        if did in self._get_metadata_provider(account=publisher_account).list_assets():
            raise DIDAlreadyExist(
                f'Asset id {did} is already registered to another asset.')

        for service in services:
            if service.type == ServiceTypes.ASSET_ACCESS:
                access_service = ServiceFactory.complete_access_service(
                    did,
                    gateway.get_access_endpoint(self._config),
                    access_service_attributes,
                    self._keeper.access_template.address,
                    self._keeper.escrow_payment_condition.address,
                    service.type
                )
                ddo.add_service(access_service)
            elif service.type == ServiceTypes.ASSET_ACCESS_PROOF:
                access_service = ServiceFactory.complete_access_service(
                    did,
                    gateway.get_access_proof_endpoint(self._config),
                    self._build_access_proof(metadata_copy, publisher_account, asset_rewards, timeout),
                    self._keeper.access_proof_template.address,
                    self._keeper.escrow_payment_condition.address,
                    service.type
                )
                ddo.add_service(access_service)
            elif service.type == ServiceTypes.NFT_ACCESS:
                access_service = ServiceFactory.complete_access_service(
                    did,
                    gateway.get_nft_access_endpoint(self._config),
                    self._build_nft_access(metadata_copy, publisher_account, asset_rewards, service.main['_numberNfts'], timeout),
                    self._keeper.nft_access_template.address,
                    self._keeper.escrow_payment_condition.address,
                    service.type
                )
                ddo.add_service(access_service)
            elif service.type == ServiceTypes.NFT_SALES:
                nft_sales_service = ServiceFactory.complete_nft_sales_service(
                    did,
                    gateway.get_access_endpoint(self._config),
                    service.attributes,
                    self._keeper.nft_sales_template.address,
                    self._keeper.escrow_payment_condition.address,
                    service.type
                )
                ddo.add_service(nft_sales_service)
            elif service.type == ServiceTypes.METADATA:
                ddo_service_endpoint = service.service_endpoint.replace('{did}', did)
                service.set_service_endpoint(ddo_service_endpoint)
                ddo.add_service(service)
            elif service.type == ServiceTypes.CLOUD_COMPUTE:
                compute_service = ServiceFactory.complete_compute_service(
                    did,
                    service.service_endpoint,
                    service.attributes,
                    self._keeper.compute_execution_condition.address,
                    self._keeper.escrow_payment_condition.address
                )
                ddo.add_service(compute_service)
            else:
                ddo.add_service(service)

        ddo.proof['signatureValue'] = self._keeper.sign_hash(
            add_ethereum_prefix_and_hash_msg(did_to_id_bytes(did)),
            publisher_account)

        # Add public key and authentication
        ddo.add_public_key(did, publisher_account.address)

        ddo.add_authentication(did, PUBLIC_KEY_TYPE_RSA)

        # Setup metadata service
        # First compute files_encrypted
        if metadata_copy['main']['type'] in ['dataset', 'algorithm']:
            assert metadata_copy['main'][
                'files'], 'files is required in the metadata main attributes.'
            logger.debug('Encrypting content urls in the metadata.')
            if not use_secret_store:
                encrypt_endpoint = gateway.get_encrypt_endpoint(self._config)
                files_encrypted = gateway.encrypt_files_dict(
                    metadata_copy['main']['files'],
                    encrypt_endpoint,
                    ddo.asset_id,
                    authorization_type)
            else:
                files_encrypted = self._get_secret_store(publisher_account) \
                    .encrypt_document(
                    did_to_id(did),
                    json.dumps(metadata_copy['main']['files']),
                )
            # only assign if the encryption worked
            if files_encrypted:
                logger.debug(f'Content urls encrypted successfully {files_encrypted}')
                index = 0
                for file in metadata_copy['main']['files']:
                    file['index'] = index
                    index = index + 1
                    del file['url']
                metadata_copy['encryptedFiles'] = files_encrypted
            else:
                raise AssertionError('Encrypting the files failed.')

        # DDO url and `Metadata` service

        logger.debug(
            f'Generated ddo and services, DID is {ddo.did},'
            f' metadata service @{ddo_service_endpoint}.')
        response = None

        # register on-chain
        if mint > 0 or royalties is not None or cap is not None:
            registered_on_chain = self._keeper.did_registry.register_mintable_did(
                did_seed,
                checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
                url=ddo_service_endpoint,
                account=publisher_account,
                cap=cap,
                royalties=royalties,
                providers=providers,
                activity_id=activity_id,
                attributes=attributes
            )
            if mint > 0:
                self._keeper.did_registry.mint(ddo.asset_id, mint, account=publisher_account)
        else:
            registered_on_chain = self._keeper.did_registry.register(
                did_seed,
                checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
                url=ddo_service_endpoint,
                account=publisher_account,
                cap=cap,
                royalties=royalties,
                providers=providers,
                activity_id=activity_id,
                attributes=attributes
            )

        if registered_on_chain is False:
            logger.warning(f'Registering {did} on-chain failed.')
            return None
        logger.info(f'Successfully registered DDO (DID={did}) on chain.')
        try:
            # publish the new ddo
            response = self._get_metadata_provider(account=publisher_account).publish_asset_ddo(ddo)
            logger.info('Asset/ddo published successfully in Metadata.')
        except ValueError as ve:
            raise ValueError(f'Invalid value to publish in the metadata: {str(ve)}')
        except Exception as e:
            logger.error(f'Publish asset in Metadata failed: {str(e)}')
        if not response:
            return None
        return ddo

    def create_compute(self, metadata, publisher_account, asset_rewards={"_amounts": [], "_receivers": []},
                       service_descriptors=None, providers=None,
                       authorization_type=ServiceAuthorizationTypes.PSK_RSA, use_secret_store=False, timeout=86400):
        """
        Register a compute to the data asset in both the keeper's DIDRegistry (on-chain) and in
        the Metadata store.

        :param metadata: dict conforming to the Metadata accepted by Nevermined Protocol.
        :param publisher_account: Account of the publisher registering this asset
        :param asset_rewards: Asset rewards distribution including amounts and receivers
        :param service_descriptors: list of ServiceDescriptor tuples of length 2.
            The first item must be one of ServiceTypes and the second
            item is a dict of parameters and values required by the service
        :param providers: list of addresses of providers of this asset (a provider is
            an ethereum account that is authorized to provide asset services)
        :param authorization_type: str indicate the authorization type that is going to be used
        to encrypt the urls.
            SecretStore, PSK-RSA and PSK-ECDSA are supported.
        :param use_secret_store: bool indicate whether to use the secret store directly for
            encrypting urls (Uses Gateway provider service if set to False)
        :return: DDO instance
        """

        metadata_copy = copy.deepcopy(metadata)
        gateway = GatewayProvider.get_gateway()

        compute_service_attributes = self._build_compute(metadata_copy, publisher_account, asset_rewards, timeout)
        service_descriptor = ServiceDescriptor.compute_service_descriptor(
            compute_service_attributes, gateway.get_execute_endpoint(self._config))

        return self.create(metadata, publisher_account, service_descriptors=[service_descriptor],
                           providers=providers, authorization_type=authorization_type,
                           use_secret_store=use_secret_store)

    def retire(self, did, account):
        """
        Retire this did of Metadata

        :param did: DID, str
        :para account: The account that owns the did
        :return: bool
        """
        try:
            ddo = self.resolve(did)
            metadata_service = ddo.get_service(ServiceTypes.METADATA)
            self._get_metadata_provider(metadata_service.service_endpoint, account).retire_asset_ddo(did)
            return True
        except MetadataGenericError as err:
            logger.error(err)
            return False

    def resolve(self, did):
        """
        When you pass a did retrieve the ddo associated.

        :param did: DID, str
        :return: DDO instance
        """
        return self._did_resolver.resolve(did)

    def search(self, text, sort=None, offset=100, page=1, metadata_url=None):
        """
        Search an asset in oceanDB using Metadata.

        :param text: String with the value that you are searching
        :param sort: Dictionary to choose order main in some value
        :param offset: Number of elements shows by page
        :param page: Page number
        :param metadata_url: Url of the Metadata where you want to search. If there is not
            provided take the default
        :return: List of assets that match with the query
        """
        assert page >= 1, f'Invalid page value {page}. Required page >= 1.'
        logger.info(f'Searching asset containing: {text}')
        return [DDO(dictionary=ddo_dict) for ddo_dict in
                self._get_metadata_provider(metadata_url).text_search(text, sort, offset, page)[
                    'results']]

    def query(self, query, sort=None, offset=100, page=1, metadata_url=None):
        """
        Search an asset in oceanDB using search query.

        :param query: dict with query parameters
            (e.g.) https://github.com/nevermined-io/metadata-api/blob/develop/docs
            /for_api_users/API.md
        :param sort: Dictionary to choose order main in some value
        :param offset: Number of elements shows by page
        :param page: Page number
        :param metadata_url: Url of the Metadata where you want to search. If there is not
            provided take the default
        :return: List of assets that match with the query.
        """
        logger.info(f'Searching asset query: {query}')
        metadata_provider = self._get_metadata_provider(metadata_url)
        return [DDO(dictionary=ddo_dict) for ddo_dict in
                metadata_provider.query_search(query, sort, offset, page)['results']]

    def order(self, did, index, consumer_account, account):
        agreement_id_seed = ServiceAgreement.create_new_agreement_id()
        agreement_id = self._keeper.agreement_manager.hash_id(agreement_id_seed, consumer_account.address)
        logger.debug(f'about to request create agreement: {agreement_id}')
        result = self._agreements.create(
            did,
            index,
            agreement_id_seed,
            consumer_account,
            account
        )
        if result is False:
            raise Exception('There was an error creating the agreement')
        return agreement_id

    def access(self, service_agreement_id, did, service_index, consumer_account,
               destination, index=None, service_type=ServiceTypes.ASSET_ACCESS):
        """
        Consume the asset data.

        Using the service endpoint defined in the ddo's service pointed to by service_definition_id.
        Consumer's permissions is checked implicitly by the secret-store during decryption
        of the contentUrls.
        The service endpoint is expected to also verify the consumer's permissions to access this
        asset.
        This method downloads and saves the asset datafiles to disk.

        :param service_type:
        :param service_agreement_id: str
        :param did: DID, str
        :param service_index: identifier of the service inside the asset DDO, str
        :param consumer_account: Account instance of the consumer
        :param destination: str path
        :param index: Index of the document that is going to be downloaded, int
        :return: str path to saved files
        """
        ddo = self.resolve(did)
        if index is not None:
            assert isinstance(index, int), logger.error('index has to be an integer.')
            assert index >= 0, logger.error('index has to be 0 or a positive integer.')
        return self._asset_consumer.access(
            service_agreement_id,
            service_index,
            ddo,
            consumer_account,
            destination,
            GatewayProvider.get_gateway(),
            self._get_secret_store(consumer_account),
            self._config,
            index,
            service_type=service_type
        )

    def download(self, did, service_index, owner_account, destination, index=None):
        """
        Download the asset data.

        Using the service endpoint defined in the ddo's service pointed to by service_definition_id.
        Consumer's permissions is checked implicitly by the secret-store during decryption
        of the contentUrls.
        The service endpoint is expected to also verify that `owner_account` is the actual owner
        of the asset.
        This method downloads and saves the asset datafiles to disk.

        :param did: DID, str
        :param service_index: identifier of the service inside the asset DDO, str
        :param owner_account: Account instance of the owner
        :param destination: str path
        :param index: Index of the document that is going to be downloaded, int
        :return: str path to saved files
        """
        ddo = self.resolve(did)
        if index is not None:
            assert isinstance(index, int), logger.error('index has to be an integer.')
            assert index >= 0, logger.error('index has to be 0 or a positive integer.')
        return self._asset_consumer.download(
            service_index,
            ddo,
            owner_account,
            destination,
            GatewayProvider.get_gateway(),
            self._get_secret_store(owner_account),
            self._config,
            index
        )

    def compute_logs(self, service_agreement_id, execution_id, account):
        """
        Get the logs of a compute workflow.

        :param service_agreement_id: The id of the service agreement that ordered the compute job, str
        :param execution_id: The id of the compute job, str
        :param account: Account instance that ordered the execution of the compute job
        :return: list, compute logs
        """
        return self._asset_consumer.compute_logs(
            service_agreement_id,
            execution_id,
            account,
            GatewayProvider.get_gateway(),
            self._config,
        )

    def compute_status(self, service_agreement_id, execution_id, account):
        """
        Get the status of a compute workflow.

        :param service_agreement_id: The id of the service agreement that ordered the compute job, str
        :param execution_id: The id of the compute job, str
        :param account: Account instance that ordered the execution of the compute job
        :return: str compute status
        """
        return self._asset_consumer.compute_status(
            service_agreement_id,
            execution_id,
            account,
            GatewayProvider.get_gateway(),
            self._config,
        )

    def owner(self, did):
        """
        Return the owner of the asset.

        :param did: DID, str
        :return: the ethereum address of the owner/publisher of given asset did, hex-str
        """
        return self._keeper.did_registry.get_did_owner(did_to_id(did))

    def owner_assets(self, owner_address):
        """
        List of Asset objects published by ownerAddress

        :param owner_address: ethereum address of owner/publisher, hex-str
        :return: list of dids
        """
        return self._keeper.did_registry.get_owner_asset_ids(owner_address)

    def consumer_assets(self, consumer_address):
        """
        List of Asset objects purchased by consumerAddress

        :param consumer_address: ethereum address of consumer, hes-str
        :return: list of dids
        """
        return self._keeper.access_condition.get_purchased_assets_by_address(
            consumer_address)

    def revoke_permissions(self, did, address_to_revoke, account):
        """
        Revoke access permission to an address.

        :param did: the id of an asset on-chain, hex str
        :param address_to_revoke: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        asset_id = add_0x_prefix(did_to_id(did))
        return self._keeper.did_registry.revoke_permission(asset_id, address_to_revoke, account)

    def get_permissions(self, did, address):
        """
        Gets access permission of a grantee

        :param did: the id of an asset on-chain, hex str
        :param address: ethereum account address, hex str
        :return: true if the address has access permission to a DID
        """
        asset_id = add_0x_prefix(did_to_id(did))
        return self._keeper.did_registry.get_permission(asset_id, address)

    def delegate_persmission(self, did, address_to_grant, account):
        """
        Grant access permission to an address.

        :param did: the id of an asset on-chain, hex str
        :param address_to_grant: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        asset_id = add_0x_prefix(did_to_id(did))
        return self._keeper.did_registry.grant_permission(asset_id, address_to_grant, account)

    def transfer_ownership(self, did, new_owner_address, account):
        """
        Transfer did ownership to an address.

        :param did: the id of an asset on-chain, hex str
        :param new_owner_address: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        asset_id = add_0x_prefix(did_to_id(did))
        return self._keeper.did_registry.transfer_did_ownership(asset_id, new_owner_address,
                                                                account)

    def execute(self, agreement_id, did, index, consumer_account, workflow_did):
        """

        :param agreement_id:representation of `bytes32` id, hexstr
        :param did: computing service did, str
        :param index: id of the service within the asset DDO, str
        :param consumer_account: Account instance of the consumer ordering the service
        :param workflow_did: the asset did (of `workflow` type) which consist of `did:nv:` and
        the assetId hex str (without `0x` prefix), str
        :return: the id of the compute job, str
        """
        workflow_ddo = self.resolve(workflow_did)
        compute_ddo = self.resolve(did)
        if index is not None:
            assert isinstance(index, int), logger.error('index has to be an integer.')
            assert index >= 0, logger.error('index has to be 0 or a positive integer.')
        return self._asset_executor.execute(
            agreement_id,
            compute_ddo,
            workflow_ddo,
            consumer_account,
            GatewayProvider.get_gateway(),
            index,
            self._config
        )

    @staticmethod
    def _build_authorization(authorization_type, public_key=None, threshold=None):
        authorization = dict()
        authorization['main'] = dict()
        authorization['main']['service'] = authorization_type
        if public_key:
            authorization['main']['publicKey'] = public_key
        if threshold:
            authorization['main']['threshold'] = threshold
        return authorization

    @staticmethod
    def _build_access(metadata, publisher_account, asset_rewards, timeout=3600):
        return {"main": {
            "name": "dataAssetAccessServiceAgreement",
            "creator": publisher_account.address,
            "price": metadata[MetadataMain.KEY]['price'],
            "timeout": timeout,
            "datePublished": metadata[MetadataMain.KEY]['dateCreated'],
            "_amounts": asset_rewards["_amounts"],
            "_receivers": asset_rewards["_receivers"],
            "_tokenAddress": asset_rewards["_tokenAddress"]
        }}

    @staticmethod
    def _build_access_proof(metadata, publisher_account, asset_rewards, timeout=3600):
        return {"main": {
            "name": "dataAssetAccessProofServiceAgreement",
            "creator": publisher_account.address,
            "price": metadata[MetadataMain.KEY]['price'],
            "timeout": timeout,
            "datePublished": metadata[MetadataMain.KEY]['dateCreated'],
            "_amounts": asset_rewards["_amounts"],
            "_hash": metadata['additionalInformation']['poseidonHash'],
            "_providerPub": [
                metadata['additionalInformation']['providerKey']['x'],
                metadata['additionalInformation']['providerKey']['y']
            ],
            "_receivers": asset_rewards["_receivers"]
        }}

    @staticmethod
    def _build_nft_access(metadata, publisher_account, asset_rewards, number_nfts, timeout=3600):
        return {"main": {
            "name": "nftAccessAgreement",
            "creator": publisher_account.address,
            "price": metadata[MetadataMain.KEY]['price'],
            "timeout": timeout,
            "datePublished": metadata[MetadataMain.KEY]['dateCreated'],
            "_amounts": asset_rewards["_amounts"],
            "_receivers": asset_rewards["_receivers"],
            "_numberNfts": number_nfts,
            "_tokenAddress": asset_rewards["_tokenAddress"]
        }}

    def _build_compute(self, metadata, publisher_account, asset_rewards, timeout=86400):
        return {"main": {
            "name": "dataAssetComputeServiceAgreement",
            "creator": publisher_account.address,
            "datePublished": metadata[MetadataMain.KEY]['dateCreated'],
            "price": metadata[MetadataMain.KEY]['price'],
            "timeout": timeout,
            "provider": self._build_provider_config(),
            "_amounts": asset_rewards["_amounts"],
            "_receivers": asset_rewards["_receivers"]
        }
        }

    @staticmethod
    def _build_provider_config():
        # TODO manage this as different class and get all the info for different services
        return {
            "type": "Azure",
            "description": "",
            "environment": {
                "cluster": {
                    "type": "Kubernetes",
                    "url": "http://10.0.0.17/xxx"
                },
                "supportedContainers": [
                    {
                        "image": "tensorflow/tensorflow",
                        "tag": "latest",
                        "checksum":
                            "sha256:cb57ecfa6ebbefd8ffc7f75c0f00e57a7fa739578a429b6f72a0df19315deadc"
                    },
                    {
                        "image": "tensorflow/tensorflow",
                        "tag": "latest",
                        "checksum":
                            "sha256:cb57ecfa6ebbefd8ffc7f75c0f00e57a7fa739578a429b6f72a0df19315deadc"
                    }
                ],
                "supportedServers": [
                    {
                        "serverId": "1",
                        "serverType": "xlsize",
                        "price": "50",
                        "cpu": "16",
                        "gpu": "0",
                        "memory": "128gb",
                        "disk": "160gb",
                        "maxExecutionTime": 86400
                    },
                    {
                        "serverId": "2",
                        "serverType": "medium",
                        "price": "10",
                        "cpu": "2",
                        "gpu": "0",
                        "memory": "8gb",
                        "disk": "80gb",
                        "maxExecutionTime": 86400
                    }
                ]
            }
        }



import os

from requests import Response

from nevermined_sdk_py.gateway.gateway import Gateway


class GatewayMock(object):
    def __init__(self, ocean_instance=None, account=None):
        if not ocean_instance:
            from tests.resources.helper_functions import get_publisher_instance
            ocean_instance = get_publisher_instance(
                init_tokens=False, use_ss_mock=False, use_gateway_mock=True
            )

        self.ocean_instance = ocean_instance
        self.account = account
        if not account:
            from tests.resources.helper_functions import get_publisher_account
            self.account = get_publisher_account()

        # nevermined_instance.agreements.watch_provider_events(self.account)

    @staticmethod
    def consume_service(did, service_agreement_id, service_endpoint, account_address, files,
                        destination_folder, *_, **__):
        for f in files:
            with open(os.path.join(destination_folder, os.path.basename(f['url'])), 'w') as of:
                of.write(f'mock data {service_agreement_id}.{service_endpoint}.{account_address}')

    @staticmethod
    def access_service(did, service_agreement_id, service_endpoint, account, destination_folder,
                       index, *_, **__):
        with open(os.path.join(destination_folder, os.path.basename(did)), 'w') as of:
            of.write(f'mock data {service_agreement_id}.{service_endpoint}.{account}')

    @staticmethod
    def encrypt_files_dict(files_dict, encrypt_endpoint, asset_id, method):
        encrypted = f'{method}.{asset_id}!!{files_dict}!!'
        return encrypted

    @staticmethod
    def execute_service(agreement_id, service_endpoint, consumer_account, workflow_ddo):
        return True

    @staticmethod
    def execute_compute_service(agreement_id, service_endpoint, consumer_account, workflow_ddo):
        r = Response()
        r._content = b'{"workflowId": "nevermined-compute-h4trr"}'
        return r

    @staticmethod
    def get_gateway_url(config):
        return Gateway.get_gateway_url(config)

    @staticmethod
    def get_access_endpoint(config):
        return f'{Gateway.get_gateway_url(config)}/services/access'

    @staticmethod
    def get_access_proof_endpoint(config):
        return f'{Gateway.get_gateway_url(config)}/services/access-proof'

    @staticmethod
    def get_consume_endpoint(config):
        return f'{Gateway.get_gateway_url(config)}/services/consume'

    @staticmethod
    def get_execute_endpoint(config):
        return f'{Gateway.get_gateway_url(config)}/services/execute'

    @staticmethod
    def get_encrypt_endpoint(config):
        return f'{Gateway.get_gateway_url(config)}/services/encrypt'

    @staticmethod
    def get_rsa_public_key(config):
        return 'rsa_public_key'

    @staticmethod
    def get_ecdsa_public_key(config):
        return 'ecdsa-public-key'



def test_token_request(publisher_instance):
    receiver_account = publisher_instance.main_account
    # Starting balance for comparison
    start_ocean = publisher_instance.accounts.balance(receiver_account).ocn

    # Make requests, assert success on request
    publisher_instance.tokens.request(receiver_account, 2000)
    # Should be no change, 2000 exceeds the max of 1000
    assert publisher_instance.accounts.balance(receiver_account).ocn == start_ocean

    amount = 500
    publisher_instance.tokens.request(receiver_account, amount)
    # Confirm balance changes
    # TODO Review representation of amounts.
    assert publisher_instance.accounts.balance(
        receiver_account).ocn == start_ocean + (amount * 1000000000000000000)


def test_transfer_tokens(publisher_instance_no_init, consumer_instance_no_init):
    receiver_account = publisher_instance_no_init.main_account
    sender_account = consumer_instance_no_init.main_account

    receiver_start_balance = publisher_instance_no_init.accounts.balance(receiver_account).ocn
    sender_start_balance = consumer_instance_no_init.accounts.balance(sender_account).ocn

    consumer_instance_no_init.tokens.transfer(receiver_account.address, 500, sender_account)

    assert publisher_instance_no_init.accounts.balance(
        receiver_account).ocn == receiver_start_balance + 500
    assert consumer_instance_no_init.accounts.balance(
        sender_account).ocn == sender_start_balance - 500



import pytest
import time

def test_get_service_endpoint(metadata_provider_instance):
    assert metadata_provider_instance.get_service_endpoint() == f'{metadata_provider_instance.url}/' + '{did}'


def test_publish_valid_ddo(asset1, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    assert metadata_provider_instance.get_asset_ddo(asset1.did)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_publish_invalid_ddo(metadata_provider_instance):
    with pytest.raises(AttributeError):
        metadata_provider_instance.publish_asset_ddo({})


def test_publish_ddo_already_registered(asset1, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    with pytest.raises(ValueError):
        metadata_provider_instance.publish_asset_ddo(asset1)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_get_asset_ddo_for_not_registered_did(metadata_provider_instance):
    invalid_did = 'did:nv:not_valid'
    result = metadata_provider_instance.get_asset_ddo(invalid_did)
    assert result == {}


def test_get_asset_metadata(asset1, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    metadata_dict = metadata_provider_instance.get_asset_metadata(asset1.did)
    assert isinstance(metadata_dict, dict)
    assert 'main' in metadata_dict
    assert 'curation' in metadata_dict
    assert 'additionalInformation' in metadata_dict
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_get_asset_metadata_for_not_registered_did(metadata_provider_instance):
    invalid_did = 'did:nv:not_valid'
    metadata_dict = metadata_provider_instance.get_asset_metadata(invalid_did)
    assert metadata_dict == {}


def test_list_assets(asset1, metadata_provider_instance):
    ddo = metadata_provider_instance.publish_asset_ddo(asset1)

    # wait for it to be searcheable
    time.sleep(3)

    assert ddo['id'] == asset1.did
    assert ddo['id'] in metadata_provider_instance.list_assets()

    assets = metadata_provider_instance.list_assets()
    assert isinstance(assets, list)
    assert isinstance(assets[0], str)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_list_assets_ddo(asset1, metadata_provider_instance):
    ddo = metadata_provider_instance.publish_asset_ddo(asset1)

    # wait for it to be searcheable
    time.sleep(3)

    assert ddo['id'] == asset1.did
    assert ddo in metadata_provider_instance.list_assets_ddo()

    assert isinstance(metadata_provider_instance.list_assets_ddo(), list)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_update_ddo(asset1, asset2, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    metadata_provider_instance.update_asset_ddo(asset1.did, asset2)
    assert metadata_provider_instance.get_asset_ddo(asset1.did).did == asset2.did
    assert metadata_provider_instance.get_asset_ddo(asset1.did).metadata['main']['name'] != \
           asset1.metadata['main'][
               'name'], 'The name has not been updated correctly.'
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_update_with_not_valifd_ddo(asset1, metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.update_asset_ddo(asset1.did, {})


def test_text_search(asset1, asset2, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    # wait for it to be searcheable
    time.sleep(3)

    result = metadata_provider_instance.text_search(text='Weather information', offset=10000)
    number_matches = result['total_results']['value']
    assert number_matches > 0

    text = 'UK'
    metadata_provider_instance.publish_asset_ddo(asset2)
    # wait for it to be searcheable
    time.sleep(3)

    result = metadata_provider_instance.text_search(text=text, offset=10000)
    number_matches = result['total_results']['value']
    assert number_matches > 0

    metadata_provider_instance.retire_asset_ddo(asset1.did)
    metadata_provider_instance.retire_asset_ddo(asset2.did)


def test_text_search_invalid_query(metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.text_search(text='', offset='Invalid')


def test_query_search(asset1, metadata_provider_instance):
    search_query = {
        "query": {
            "bool": {
                "must": [{
                    "match": {"service.attributes.main.type": "dataset"}
                }]
            }
        }
    }

    metadata_provider_instance.publish_asset_ddo(asset1)
    # wait for it to be searcheable
    time.sleep(3)

    result = metadata_provider_instance.query_search(search_query=search_query, offset=10000)
    number_matches = result['total_results']['value']
    assert number_matches > 0

    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_query_search_invalid_query(metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.query_search(search_query='')


def test_retire_ddo(asset1, metadata_provider_instance):
    n = len(metadata_provider_instance.list_assets())
    metadata_provider_instance.publish_asset_ddo(asset1)

    # wait for elasticsearch
    time.sleep(3)

    response = metadata_provider_instance.retire_asset_ddo(asset1.did)
    assert response.ok is True

def test_retire_not_published_did(metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.retire_asset_ddo('did:nv:not_registered')



import logging

from common_utils_py.data_store.storage_base import StorageBase

logger = logging.getLogger(__name__)


class AuthTokensStorage(StorageBase):
    AUTH_TOKENS_TABLE = 'auth_tokens'

    def write_token(self, address, signed_token, created_at):
        """
        Store signed token for session management.

        :param address: hex str the ethereum address that signed the token
        :param signed_token: hex str the signed token
        :param created_at: date-time of token creation
        """
        logger.debug(f'Writing token to `auth_tokens` storage: '
                     f'account={address}, token={signed_token}')
        self._run_query(
            f'''CREATE TABLE IF NOT EXISTS {self.AUTH_TOKENS_TABLE}
               (address VARCHAR PRIMARY KEY, signed_token VARCHAR, created VARCHAR);'''
        )
        self._run_query(
            f'''INSERT OR REPLACE
                INTO {self.AUTH_TOKENS_TABLE}
                VALUES (?,?,?)''',
            [address, signed_token, created_at],
        )

    def update_token(self, address, signed_token, created_at):
        """
        Update/replace the stored signed token for the given ethereum address

        :param address: hex str the ethereum address that signed the token
        :param signed_token: hex str the signed token
        :param created_at: date-time of token creation
        """
        logger.debug(f'Updating token already in `auth_tokens` storage: '
                     f'account={address}, token={signed_token}')
        self._run_query(
            f'''UPDATE {self.AUTH_TOKENS_TABLE}
                SET signed_token=?, created=?
                WHERE address=?''',
            (signed_token, created_at, address),
        )

    def read_token(self, address):
        """
        Retrieve stored signed token for the given ethereum address

        :param address: hex str the ethereum address that signed the token
        :return: tuple (signed_token, created_at)
        """
        try:
            rows = [row for row in self._run_query(
                f'''SELECT signed_token, created
                    FROM {self.AUTH_TOKENS_TABLE}
                    WHERE address=?;''',
                (address,))
                    ]
            token, timestamp = rows[0] if rows else (None, None)
            logger.debug(f'Read auth token from `auth_tokens` storage: '
                         f'account={address}, token={token}')
            return token, timestamp

        except Exception as e:
            logging.error(f'Error reading token: {e}')
            return None, None



from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from nevermined_sdk_py.nevermined.auth import Auth
from tests.resources.helper_functions import get_publisher_account


def test_get_token():
    auth = Auth(Keeper.get_instance(), ':memory:')
    acc = get_publisher_account()
    token = auth.get(acc)
    assert isinstance(token, str), 'Invalid auth token type.'
    assert token.startswith('0x'), 'Invalid auth token.'
    parts = token.split('-')
    assert len(parts) == 2, 'Invalid token, timestamp separator is not found.'

    address = auth.check(token)
    assert address != '0x0', 'Verifying token failed.'


def test_check_token(web3_instance):
    auth = Auth(Keeper.get_instance(), ':memory:')
    acc = get_publisher_account()

    token = auth.get(acc)
    address = auth.check(token)
    assert address != '0x0', 'Verifying token failed.'

    sig = token.split('-')[0]
    assert auth.check(sig) == '0x0'

    # Test token expiration


def test_store_token():
    auth = Auth(Keeper.get_instance(), ':memory:')
    acc = get_publisher_account()
    token = auth.store(acc)
    assert auth.check(token) == acc.address, 'invalid token, check failed.'
    # verify it is saved
    assert auth.restore(acc) == token, 'Restoring token failed.'


def test_restore_token(publisher_instance):
    auth = Auth(Keeper.get_instance(), ':memory:')
    acc = get_publisher_account()
    assert auth.restore(acc) is None, 'Expecting None when restoring non-existing token.'

    token = auth.store(acc)
    assert auth.check(token) == acc.address, 'invalid token, check failed.'
    # verify it is saved
    assert auth.restore(acc) == token, 'Restoring token failed.'



class ConfigProvider:
    """Provides the Config instance."""
    _config = None

    @staticmethod
    def get_config():
        """ Get a Config instance."""
        assert ConfigProvider._config, 'set_config first.'
        return ConfigProvider._config

    @staticmethod
    def set_config(config):
        """
         Set a Config instance.

        :param config: Config
        :return:  New Config instance.
        """
        ConfigProvider._config = config



def test_nevermined_instance(publisher_instance):
    assert publisher_instance.tokens
    assert publisher_instance.agreements
    assert publisher_instance.assets
    assert publisher_instance.accounts
    assert publisher_instance.services
    assert publisher_instance.auth
    assert publisher_instance.templates
    assert publisher_instance.secret_store



class Providers:
    """Nevermined assets class."""

    def __init__(self, keeper, did_resolver, config):
        self._keeper = keeper
        self._did_resolver = did_resolver
        self._config = config

    def add(self, did, provider_address, account):
        return self._keeper.did_registry.add_provider(did, provider_address, account)

    def remove(self, did, provider_address, account):
        return self._keeper.did_registry.remove_provider(did, provider_address, account)

    def list(self, did):
        return self._keeper.did_registry.get_did_providers(did)



import configparser
import logging
import os
import site
from pathlib import Path

import nevermined_contracts

DEFAULT_KEEPER_HOST = 'localhost'
DEFAULT_KEEPER_PORT = 8545
DEFAULT_KEEPER_URL = 'http://localhost:8545'
DEFAULT_KEEPER_PATH = nevermined_contracts.get_artifacts_path()
DEFAULT_GAS_LIMIT = 4000000
DEFAULT_NAME_METADATA_URL = 'http://localhost:5000'
DEFAULT_NAME_GATEWAY_URL = 'http://localhost:8030'
DEFAULT_STORAGE_PATH = 'sdk.db'

NAME_KEEPER_URL = 'keeper.url'
NAME_KEEPER_PATH = 'keeper.path'
NAME_GAS_LIMIT = 'gas_limit'
NAME_METADATA_URL = 'metadata.url'
NAME_GATEWAY_URL = 'gateway.url'
NAME_STORAGE_PATH = 'storage.path'
NAME_AUTH_TOKEN_MESSAGE = 'auth_token_message'
NAME_AUTH_TOKEN_EXPIRATION = 'auth_token_expiration'

NAME_SECRET_STORE_URL = 'secret_store.url'
NAME_PARITY_URL = 'parity.url'
NAME_PARITY_ADDRESS = 'parity.address'
NAME_PARITY_PASSWORD = 'parity.password'

environ_names = {
    NAME_KEEPER_URL: ['KEEPER_URL', 'Keeper URL'],
    NAME_KEEPER_PATH: ['KEEPER_PATH', 'Path to the keeper contracts'],
    NAME_GAS_LIMIT: ['GAS_LIMIT', 'Gas limit'],
    NAME_METADATA_URL: ['METADATA_URL', 'Metadata URL'],
    NAME_GATEWAY_URL: ['GATEWAY_URL', 'Gateway URL'],
    NAME_STORAGE_PATH: ['STORAGE_PATH', 'Path to the local database file'],
    NAME_AUTH_TOKEN_MESSAGE: ['AUTH_TOKEN_MESSAGE',
                              'Message to use for generating user auth token'],
    NAME_AUTH_TOKEN_EXPIRATION: ['AUTH_TOKEN_EXPIRATION',
                                 'Auth token expiration time expressed in seconds'],
    NAME_SECRET_STORE_URL: ['SECRET_STORE_URL', 'Secret Store URL'],
    NAME_PARITY_URL: ['PARITY_URL', 'Parity URL'],
    NAME_PARITY_ADDRESS: ['PARITY_ADDRESS', 'Parity address'],
    NAME_PARITY_PASSWORD: ['PARITY_PASSWORD', 'Parity password'],
}

config_defaults = {
    'keeper-contracts': {
        NAME_KEEPER_URL: DEFAULT_KEEPER_URL,
        NAME_KEEPER_PATH: DEFAULT_KEEPER_PATH,
        NAME_GAS_LIMIT: DEFAULT_GAS_LIMIT,
        NAME_SECRET_STORE_URL: '',
        NAME_PARITY_URL: '',
        NAME_PARITY_ADDRESS: '',
        NAME_PARITY_PASSWORD: '',
    },
    'resources': {
        NAME_METADATA_URL: DEFAULT_NAME_METADATA_URL,
        NAME_GATEWAY_URL: DEFAULT_NAME_GATEWAY_URL,
        NAME_STORAGE_PATH: DEFAULT_STORAGE_PATH,
        NAME_AUTH_TOKEN_MESSAGE: '',
        NAME_AUTH_TOKEN_EXPIRATION: ''
    }
}


class Config(configparser.ConfigParser):
    """Class to manage the nevermined-sdk-py configuration."""

    def __init__(self, filename=None, options_dict=None, **kwargs):
        """
        Initialize Config class.

        Options available:

        [keeper-contracts]
        keeper.url = http://localhost:8545                            # Keeper-contracts url.
        keeper.path = artifacts                                       # Path of json abis.
        secret_store.url = http://localhost:12001                     # Secret store url.
        parity.url = http://localhost:8545                            # Parity client url.
        parity.address = 0x00bd138abd70e2f00903268f3db08f2d25677c9e   # Partity account address.
        parity.password = node0                                       # Parity account password.
        [resources]
        metadata.url = http://localhost:5000                          # Metadata url.
        gateway.url = http://localhost:8030                           # Gateway url.
        storage.path = sdk.db                                    # Path of sla back-up storage.

        :param filename: Path of the config file, str.
        :param options_dict: Python dict with the config, dict.
        :param kwargs: Additional args. If you pass text, you have to pass the plain text
        configuration.
        """
        configparser.ConfigParser.__init__(self)

        self.read_dict(config_defaults)
        self._web3_provider = None
        self._section_name = 'keeper-contracts'
        self._logger = logging.getLogger('config')

        if filename:
            self._logger.debug(f'Config: loading config file {filename}')
            with open(filename) as fp:
                text = fp.read()
                self.read_string(text)
        else:
            if 'text' in kwargs:
                self.read_string(kwargs['text'])

        if options_dict:
            self._logger.debug(f'Config: loading from dict {options_dict}')
            self.read_dict(options_dict)

        self._load_environ()

    def _load_environ(self):
        for option_name, environ_item in environ_names.items():
            value = os.environ.get(environ_item[0])
            if value is not None:
                self._logger.debug(f'Config: setting environ {option_name} = {value}')
                self.set(self._section_name, option_name, value)

    @property
    def keeper_path(self):
        """Path where the keeper-contracts artifacts are allocated."""
        keeper_path_string = self.get(self._section_name, NAME_KEEPER_PATH)
        path = Path(keeper_path_string).expanduser().resolve()

        if os.path.exists(path):
            pass
        elif os.getenv('VIRTUAL_ENV'):
            path = os.path.join(os.getenv('VIRTUAL_ENV'), 'artifacts')
        else:
            path = os.path.join(site.PREFIXES[0], 'artifacts')
        return path

    @property
    def storage_path(self):
        """Path to save the current execution of the service agreements and restart if needed."""
        return self.get('resources', NAME_STORAGE_PATH)

    @property
    def keeper_url(self):
        """URL of the keeper. (e.g.): http://mykeeper:8545."""
        return self.get(self._section_name, NAME_KEEPER_URL)

    @property
    def gas_limit(self):
        """Ethereum gas limit."""
        return int(self.get(self._section_name, NAME_GAS_LIMIT))

    @property
    def metadata_url(self):
        """URL of metadata component. (e.g.): http://mymetadata:5000."""
        return self.get('resources', NAME_METADATA_URL)

    @property
    def gateway_url(self):
        """URL of gateway component. (e.g.): http://mygateway:8030."""
        return self.get('resources', NAME_GATEWAY_URL)

    @property
    def secret_store_url(self):
        """URL of the secret store component. (e.g.): http://mysecretstore:12001."""
        return self.get(self._section_name, NAME_SECRET_STORE_URL)

    @property
    def parity_url(self):
        """URL of parity client. (e.g.): http://myparity:8545."""
        return self.get(self._section_name, NAME_PARITY_URL)

    @property
    def parity_address(self):
        """Parity address. (e.g.): 0x00bd138abd70e2f00903268f3db08f2d25677c9e."""
        return self.get(self._section_name, NAME_PARITY_ADDRESS)

    @property
    def parity_password(self):
        """Parity password for your address. (e.g.): Mypass."""
        return self.get(self._section_name, NAME_PARITY_PASSWORD)

    @property
    def downloads_path(self):
        """Path for the downloads of assets."""
        return self.get('resources', 'downloads.path')

    @property
    def auth_token_message(self):
        return self.get('resources', NAME_AUTH_TOKEN_MESSAGE)

    @property
    def auth_token_expiration(self):
        return self.get('resources', NAME_AUTH_TOKEN_EXPIRATION)

    @property
    def web3_provider(self):
        """Web3 provider"""
        return self._web3_provider

    @web3_provider.setter
    def web3_provider(self, web3_provider):
        self._web3_provider = web3_provider



import hashlib
import secrets

from tests.resources.helper_functions import get_assset_json_text


def test_secret_store(publisher_instance):
    metadata_json = get_assset_json_text(
        'https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs'
        '/examples/access/v0.1/ddo1.json')
    document_id = hashlib.sha256((metadata_json + secrets.token_hex(32)).encode()).hexdigest()
    publisher_account = publisher_instance.main_account

    encrypt_content = publisher_instance.secret_store.encrypt(document_id, metadata_json,
                                                              publisher_account)

    assert metadata_json == publisher_instance.secret_store.decrypt(document_id,
                                                                    encrypt_content,
                                                                    publisher_account)



from nevermined_sdk_py.gateway.gateway import Gateway


class GatewayProvider(object):
    """Provides a Gateway instance."""
    _gateway_class = Gateway

    @staticmethod
    def get_gateway():
        """ Get a Gateway instance."""
        return GatewayProvider._gateway_class()

    @staticmethod
    def set_gateway_class(gateway_class):
        """
         Set a Gateway class.

        :param gateway_class: Gateway-compatible class
        """
        GatewayProvider._gateway_class = gateway_class



import hashlib
import secrets
from unittest.mock import MagicMock, Mock

from contracts_lib_py.account import Account

from nevermined_sdk_py import ConfigProvider
from nevermined_sdk_py.secret_store.secret_store import SecretStore
from tests.resources.helper_functions import get_assset_json_text


def test_secret_store_encrypt_decrypt():
    metadata_json = get_assset_json_text(
        'https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs'
        '/examples/access/v0.1/ddo1.json')
    document_id = hashlib.sha256((metadata_json + secrets.token_hex(32)).encode()).hexdigest()
    print(document_id)
    config = ConfigProvider.get_config()
    ss_client = Mock
    ss_client.publish_document = MagicMock(return_value='!!document!!')
    ss_client.decrypt_document = MagicMock(return_value=metadata_json)
    SecretStore.set_client(ss_client)

    ss_args = (config.secret_store_url, config.parity_url, Account('0x0000', 'aaa'))
    result = SecretStore(*ss_args).encrypt_document(document_id, metadata_json)
    print(result)
    assert SecretStore(*ss_args).decrypt_document(document_id, result) == metadata_json


def test_set_secret_store_url():
    ss_client = Mock
    SecretStore.set_client(ss_client)
    SecretStore.set_secret_store_url(ss_client, url='http://secret-store-changed-url')
    assert ss_client._secret_store_url == 'http://secret-store-changed-url'


def test_set_client():
    pass



import logging

from nevermined_sdk_py.gateway import GatewayProvider
from common_utils_py.did_resolver.did_resolver import DIDResolver
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper


logger = logging.getLogger(__name__)


def execute_computation(event, agreement_id, did, service_agreement, consumer_account,
                        consume_callback, workflow_did):
    """
    Consumption of an asset after get the event call.

    :param event: AttributeDict with the event data.
    :param agreement_id: id of the agreement, hex str
    :param did: DID, str
    :param service_agreement: ServiceAgreement instance
    :param consumer_account: Account instance of the consumer
    :param consume_callback:
    :param secret_store_url: str URL of secret store node for retrieving decryption keys
    :param parity_url: str URL of parity client to use for secret store encrypt/decrypt
    :param downloads_path: str path to save downloaded files
    """
    logger.debug(f"consuming asset (agreementId {agreement_id}) after event {event}.")
    if consume_callback:
        gateway = GatewayProvider.get_gateway()
        consume_callback(
            agreement_id,
            DIDResolver(Keeper.get_instance().did_registry).resolve(did),
            DIDResolver(Keeper.get_instance().did_registry).resolve(workflow_did),
            consumer_account,
            gateway
        )



from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes, ServiceTypesIndices
from common_utils_py.utils.utilities import to_checksum_addresses

from nevermined_sdk_py.gateway.gateway import Gateway
from nevermined_sdk_py.nevermined.agreements import check_token_address
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from tests.resources.helper_functions import log_event
from tests.resources.mocks.gateway_mock import GatewayMock


def test_sign_agreement(publisher_instance, consumer_instance, registered_ddo):
    # point consumer_instance's Gateway mock to the publisher's nevermined instance
    Gateway.set_http_client(
        GatewayMock(publisher_instance, publisher_instance.main_account))

    consumer_acc = consumer_instance.main_account
    keeper = Keeper.get_instance()

    publisher_acc = publisher_instance.main_account

    did = registered_ddo.did
    asset_id = registered_ddo.asset_id
    ddo = consumer_instance.assets.resolve(did)
    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS, ddo)

    price = service_agreement.get_price()

    # Give consumer some tokens
    keeper.dispenser.request_tokens(100, consumer_acc)

    agreement_id_seed = ServiceAgreement.create_new_agreement_id()
    agreement_id = keeper.agreement_manager.hash_id(agreement_id_seed, consumer_acc.address)

    success = publisher_instance.agreements.create(
        did,
        ServiceTypesIndices.DEFAULT_ACCESS_INDEX,
        agreement_id_seed,
        consumer_acc,
        consumer_acc
    )
    assert success, 'createAgreement failed.'

    # Verify condition types (condition contracts)
    agreement_values = keeper.agreement_manager.get_agreement(agreement_id)
    assert agreement_values.did == asset_id, ''
    cond_types = keeper.access_template.get_condition_types()
    for i, cond_id in enumerate(agreement_values.condition_ids):
        cond = keeper.condition_manager.get_condition(cond_id)
        assert cond.type_ref == cond_types[i]

    access_cond_id, lock_cond_id, escrow_cond_id = agreement_values.condition_ids

    # Fulfill lock_payment_condition is done automatically when create agreement is done correctly
    assert 2 == keeper.condition_manager.get_condition_state(lock_cond_id), ''
    assert 1 == keeper.condition_manager.get_condition_state(access_cond_id), ''
    assert 1 == keeper.condition_manager.get_condition_state(escrow_cond_id), ''

    # Fulfill access_condition
    tx_hash = keeper.access_condition.fulfill(
        agreement_id, asset_id, consumer_acc.address, publisher_acc
    )
    keeper.access_condition.get_tx_receipt(tx_hash)
    assert 2 == keeper.condition_manager.get_condition_state(access_cond_id), ''
    event = keeper.access_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.access_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessSecretStoreCondition.Fulfilled'

    # Fulfill escrow_payment_condition

    amounts = service_agreement.get_amounts_int()
    receivers = service_agreement.get_receivers()
    token_address = check_token_address(
        keeper, service_agreement.get_param_value_by_name('_tokenAddress'))

    tx_hash = keeper.escrow_payment_condition.fulfill(
        agreement_id, asset_id, amounts, receivers, consumer_acc.address,
        keeper.escrow_payment_condition.address, token_address, lock_cond_id,
        access_cond_id, publisher_acc
    )
    keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
    assert 2 == keeper.condition_manager.get_condition_state(escrow_cond_id), ''
    event = keeper.escrow_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.escrow_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for EscrowReward.Fulfilled'
    publisher_instance.assets.retire(did, publisher_acc)


def test_agreement_status(setup_agreements_environment, agreements):
    (
        keeper,
        publisher_acc,
        consumer_acc,
        agreement_id_seed,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),

    ) = setup_agreements_environment

    # Give consumer some tokens
    keeper.dispenser.request_tokens(100, consumer_acc)

    success = keeper.access_template.create_agreement(
        agreement_id_seed,
        asset_id,
        [access_cond_id[0], lock_cond_id[0], escrow_cond_id[0]],
        service_agreement.conditions_timelocks,
        service_agreement.conditions_timeouts,
        consumer_acc.address,
        consumer_acc
    )
    print('create agreement: ', success)
    print('--- agreement_id', agreement_id)
    assert success, f'createAgreement failed {success}'
    event = keeper.access_template.subscribe_agreement_created(
        agreement_id,
        10,
        log_event(keeper.access_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 1,
                                                              "accessSecretStore": 1,
                                                              "escrowReward": 1
                                                              }
                                               }

    amounts = service_agreement.get_amounts_int()
    receivers = service_agreement.get_receivers()

    token_address = keeper.token.address

    agreements.conditions.lock_payment(
        agreement_id, asset_id, amounts, receivers, token_address, consumer_acc)

    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for LockRewardCondition.Fulfilled'
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 2,
                                                              "accessSecretStore": 1,
                                                              "escrowReward": 1
                                                              }
                                               }
    tx_hash = keeper.access_condition.fulfill(
        agreement_id, asset_id, consumer_acc.address, publisher_acc)
    keeper.access_condition.get_tx_receipt(tx_hash)
    event = keeper.access_condition.subscribe_condition_fulfilled(
        agreement_id,
        20,
        log_event(keeper.access_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessSecretStoreCondition.Fulfilled'
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 2,
                                                              "accessSecretStore": 2,
                                                              "escrowReward": 1
                                                              }
                                               }

    tx_hash = keeper.escrow_payment_condition.fulfill(
        agreement_id, asset_id, amounts, receivers, consumer_acc.address,
        keeper.escrow_payment_condition.address, token_address, lock_cond_id[1],
        access_cond_id[1], publisher_acc
    )
    keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
    event = keeper.escrow_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.escrow_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for EscrowReward.Fulfilled'
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 2,
                                                              "accessSecretStore": 2,
                                                              "escrowReward": 2
                                                              }
                                               }



import logging


class Tokens:
    """Nevermined token class."""

    def __init__(self, keeper):
        self._keeper = keeper

    def request(self, account, amount):
        """
        Request a number of tokens to be minted and transfered to `account`

        :param account: Account instance requesting the tokens
        :param amount: int number of tokens to request
        :return: bool
        """
        if not self._keeper.dispenser:
            network_id = self._keeper.get_network_id()
            logging.warning(f'The Dispenser contract is not available for the current '
                            f'keeper network {self._keeper.get_network_name(network_id)}.')
            return False

        return self._keeper.dispenser.request_tokens(amount, account)

    def transfer(self, receiver_address, amount, sender_account):
        """
        Transfer a number of tokens from `sender_account` to `receiver_address`

        :param receiver_address: hex str ethereum address to receive this transfer of tokens
        :param amount: int number of tokens to transfer
        :param sender_account: Account instance to take the tokens from
        :return: bool
        """
        self._keeper.token.transfer(receiver_address, amount, sender_account)



import logging
from time import sleep
from contracts_lib_py.keeper import Keeper

import pytest
from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_factory import ServiceDescriptor
from common_utils_py.agreements.service_types import ServiceTypes, ServiceTypesIndices
from common_utils_py.did import DID, did_to_id
from contracts_lib_py.exceptions import DIDNotFound
from contracts_lib_py.web3_provider import Web3Provider
from eth_utils import add_0x_prefix

from tests.resources.helper_functions import log_event


def create_asset(publisher_instance, ddo_sample):
    nevermined = publisher_instance
    acct = nevermined.main_account
    asset = ddo_sample
    my_secret_store = 'http://myownsecretstore.com'
    auth_service = ServiceDescriptor.authorization_service_descriptor(
        {'main': {'service': 'SecretStore', 'publicKey': '0casd',
                  'threshold': '1'}}, my_secret_store)
    asset_rewards = {
        "_amounts": ["10", "2"],
        "_receivers": ["0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e", "0x068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0"]
    }
    return nevermined.assets.create(asset.metadata, acct, [auth_service], asset_rewards=asset_rewards)


def test_register_asset(publisher_instance, ddo_sample):
    logging.debug("".format())
    ##########################################################
    # Setup account
    ##########################################################
    publisher = publisher_instance.main_account

    # ensure Nevermined token balance
    if publisher_instance.accounts.balance(publisher).ocn == 0:
        publisher_instance.accounts.request_tokens(publisher, 200)

    # You will need some token to make this transfer!
    assert publisher_instance.accounts.balance(publisher).ocn > 0

    ##########################################################
    # Create an asset DDO with valid metadata
    ##########################################################
    asset = ddo_sample

    ##########################################################
    # Register using high-level interface
    ##########################################################
    ddo = publisher_instance.assets.create(asset.metadata, publisher)
    publisher_instance.assets.retire(ddo.did, publisher)


def test_resolve_did(publisher_instance, metadata):
    # prep ddo
    # metadata = Metadata.get_example()
    publisher = publisher_instance.main_account
    # happy path
    original_ddo = publisher_instance.assets.create(metadata, publisher, activity_id=publisher.address, attributes='test123')
    did = original_ddo.did
    ddo = publisher_instance.assets.resolve(did).as_dictionary()
    original = original_ddo.as_dictionary()
    assert ddo['publicKey'] == original['publicKey']
    assert ddo['authentication'] == original['authentication']
    assert ddo['service']
    assert original['service']
    metadata = ddo['service'][0]['attributes']
    if 'datePublished' in metadata['main']:
        metadata['main'].pop('datePublished')
    assert ddo['service'][0]['attributes']['main']['name'] == \
           original['service'][0]['attributes']['main']['name']
    assert ddo['service'][1] == original['service'][1]

    # Can't resolve unregistered asset
    unregistered_did = DID.encoded_did({"0": "0x00112233445566"})
    with pytest.raises(DIDNotFound):
        publisher_instance.assets.resolve(unregistered_did)

    # Raise error on bad did
    invalid_did = "did:nv:0123456789"
    with pytest.raises(DIDNotFound):
        publisher_instance.assets.resolve(invalid_did)
    publisher_instance.assets.retire(did, publisher)


def test_create_data_asset(publisher_instance, consumer_instance, ddo_sample):
    """
    Setup accounts and asset, register this asset on Nevermined Metadata.
    """

    logging.debug("".format())
    ##########################################################
    # Setup 2 accounts
    ##########################################################
    publisher_acct = publisher_instance.main_account
    consumer_acct = consumer_instance.main_account

    # ensure Nevermined token balance
    if publisher_instance.accounts.balance(publisher_acct).ocn == 0:
        rcpt = publisher_instance.accounts.request_tokens(publisher_acct, 200)
        Web3Provider.get_web3().eth.waitForTransactionReceipt(rcpt)
    if consumer_instance.accounts.balance(consumer_acct).ocn == 0:
        rcpt = consumer_instance.accounts.request_tokens(consumer_acct, 200)
        Web3Provider.get_web3().eth.waitForTransactionReceipt(rcpt)

    # You will need some token to make this transfer!
    assert publisher_instance.accounts.balance(publisher_acct).ocn > 0
    assert consumer_instance.accounts.balance(consumer_acct).ocn > 0

    ##########################################################
    # Create an Asset with valid metadata
    ##########################################################
    asset = ddo_sample

    ##########################################################
    # List currently published assets
    ##########################################################
    meta_data_assets = publisher_instance.assets.search('')
    if meta_data_assets:
        print("Currently registered assets:")
        print(meta_data_assets)

    if asset.did in meta_data_assets:
        publisher_instance.assets.resolve(asset.did)
        publisher_instance.assets.retire(asset.did, publisher_acct)
    # Publish the metadata
    new_asset = publisher_instance.assets.create(asset.metadata, publisher_acct)

    # get_asset_metadata only returns 'main' key, is this correct?
    published_metadata = consumer_instance.assets.resolve(new_asset.did)

    assert published_metadata
    # only compare top level keys
    assert sorted(list(asset.metadata['main'].keys())).remove('files') == sorted(
        list(published_metadata.metadata.keys())).remove('encryptedFiles')
    publisher_instance.assets.retire(new_asset.did, publisher_acct)


def test_create_asset_with_different_secret_store(publisher_instance, ddo_sample):
    acct = publisher_instance.main_account

    asset = ddo_sample
    my_secret_store = 'http://myownsecretstore.com'
    auth_service = ServiceDescriptor.authorization_service_descriptor(
        {'main': {'service': 'SecretStore', 'publicKey': '0casd',
                  'threshold': '1'}}, my_secret_store)
    new_asset = publisher_instance.assets.create(asset.metadata, acct, [auth_service])
    assert new_asset.get_service(ServiceTypes.AUTHORIZATION).service_endpoint == my_secret_store
    assert new_asset.get_service(ServiceTypes.ASSET_ACCESS)
    assert new_asset.get_service(ServiceTypes.METADATA)
    publisher_instance.assets.retire(new_asset.did, acct)

    access_service = ServiceDescriptor.access_service_descriptor(
        {"main": {
            "name": "dataAssetAccessServiceAgreement",
            "creator": '0x1234',
            "price": '1',
            "timeout": 3600,
            "datePublished": '2019-08-30T12:19:54Z'
        }}, ''
    )
    new_asset = publisher_instance.assets.create(asset.metadata, acct, [access_service])
    assert new_asset.get_service(ServiceTypes.AUTHORIZATION)
    assert new_asset.get_service(ServiceTypes.ASSET_ACCESS)
    assert new_asset.get_service(ServiceTypes.METADATA)
    publisher_instance.assets.retire(new_asset.did, acct)

    new_asset = publisher_instance.assets.create(asset.metadata, acct)
    assert new_asset.get_service(ServiceTypes.AUTHORIZATION)
    assert new_asset.get_service(ServiceTypes.ASSET_ACCESS)
    assert new_asset.get_service(ServiceTypes.METADATA)
    publisher_instance.assets.retire(new_asset.did, acct)


def test_asset_owner(publisher_instance, ddo_sample):
    acct = publisher_instance.main_account

    asset = ddo_sample
    my_secret_store = 'http://myownsecretstore.com'
    auth_service = ServiceDescriptor.authorization_service_descriptor(
        {'main': {'service': 'SecretStore', 'publicKey': '0casd',
                  'threshold': '1'}}, my_secret_store)
    new_asset = publisher_instance.assets.create(asset.metadata, acct, [auth_service])

    assert publisher_instance.assets.owner(new_asset.did) == acct.address
    publisher_instance.assets.retire(new_asset.did, acct)


def test_owner_assets(publisher_instance, ddo_sample):
    acct = publisher_instance.main_account
    assets_owned = len(publisher_instance.assets.owner_assets(acct.address))
    asset = create_asset(publisher_instance, ddo_sample)
    assert len(publisher_instance.assets.owner_assets(acct.address)) == assets_owned + 1
    publisher_instance.assets.retire(asset.did, acct)


def test_assets_consumed(publisher_instance, consumer_instance, ddo_sample):
    acct = consumer_instance.main_account
    consumed_assets = len(publisher_instance.assets.consumer_assets(acct.address))
    asset = create_asset(publisher_instance, ddo_sample)
    service = asset.get_service(service_type=ServiceTypes.ASSET_ACCESS)
    service_dict = service.as_dictionary()
    sa = ServiceAgreement.from_service_dict(service_dict)
    keeper = publisher_instance.keeper

    def grant_access(event, instance, agr_id, did, cons_address, account):
        instance.agreements.conditions.grant_access(
            agr_id, add_0x_prefix(did_to_id(did)), cons_address, account)

    agreement_id = consumer_instance.assets.order(
        asset.did, sa.index, acct, acct)
    keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        15,
        grant_access,
        (publisher_instance, agreement_id, asset.did,
         acct.address, publisher_instance.main_account),
        wait=True
    )

    keeper.access_condition.subscribe_condition_fulfilled(
        agreement_id,
        15,
        log_event(keeper.access_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert publisher_instance.agreements.is_access_granted(agreement_id, asset.did, acct.address)

    assert len(publisher_instance.assets.consumer_assets(acct.address)) == consumed_assets + 1
    publisher_instance.assets.retire(asset.did, acct)


def test_assets_resolve(publisher_instance, metadata):
    publisher = publisher_instance.main_account
    ddo = publisher_instance.assets.create(metadata, publisher)
    ddo_resolved = publisher_instance.assets.resolve(ddo.did)
    assert ddo.did == ddo_resolved.did
    publisher_instance.assets.retire(ddo.did, publisher)


def test_assets_search(publisher_instance, metadata):
    publisher = publisher_instance.main_account
    ddo = publisher_instance.assets.create(metadata, publisher)
    sleep(3)
    assert len(publisher_instance.assets.search('Monkey')) > 0
    publisher_instance.assets.retire(ddo.did, publisher)


def test_assets_algorithm(publisher_instance, algorithm_ddo):
    # Allow publish an algorithm
    publisher = publisher_instance.main_account
    metadata = algorithm_ddo['service'][0]
    ddo = publisher_instance.assets.create(metadata['attributes'], publisher)
    assert ddo
    publisher_instance.assets.retire(ddo.did, publisher)


def test_assets_workflow(publisher_instance, workflow_ddo):
    # Allow publish an workflow
    publisher = publisher_instance.main_account
    metadata = workflow_ddo['service'][0]
    ddo = publisher_instance.assets.create(metadata['attributes'], publisher)
    assert ddo
    publisher_instance.assets.retire(ddo.did, publisher)


def test_assets_compute(publisher_instance, metadata):
    publisher = publisher_instance.main_account
    asset_rewards = {
        "_amounts": ["10", "2"],
        "_receivers": ["0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e", "0x068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0"]
    }
    ddo = publisher_instance.assets.create_compute(metadata, publisher, asset_rewards=asset_rewards)
    assert ddo
    publisher_instance.assets.retire(ddo.did, publisher)


def test_transfer_ownership(publisher_instance, metadata, consumer_instance):
    publisher = publisher_instance.main_account
    consumer = consumer_instance.main_account
    ddo = publisher_instance.assets.create(metadata, publisher)
    owner = publisher_instance.assets.owner(ddo.did)
    assert owner == publisher.address
    publisher_instance.assets.transfer_ownership(ddo.did, consumer.address, publisher)
    assert publisher_instance.assets.owner(ddo.did) == consumer.address
    publisher_instance.assets.retire(ddo.did, publisher)


def test_grant_permissions(publisher_instance, metadata, consumer_instance):
    publisher = publisher_instance.main_account
    consumer = consumer_instance.main_account
    ddo = publisher_instance.assets.create(metadata, publisher)

    assert not publisher_instance.assets.get_permissions(ddo.did, consumer.address)
    publisher_instance.assets.delegate_persmission(ddo.did, consumer.address, publisher)
    assert publisher_instance.assets.get_permissions(ddo.did, consumer.address)
    publisher_instance.assets.revoke_permissions(ddo.did, consumer.address, publisher)
    assert not publisher_instance.assets.get_permissions(ddo.did, consumer.address)

    publisher_instance.assets.retire(ddo.did, publisher)

@pytest.mark.skip(reason='Compute stack disabled for now. See https://github.com/nevermined-io/internal/issues/218')
def test_execute_workflow(publisher_instance_no_init, consumer_instance_no_init, metadata, algorithm_ddo, workflow_ddo):
    consumer = publisher_instance_no_init.main_account
    publisher = consumer_instance_no_init.main_account

    # publish compute
    ddo_computing = publisher_instance_no_init.assets.create_compute(metadata, publisher)

    # publish algorithm
    metadata = algorithm_ddo['service'][0]
    ddo_algorithm = consumer_instance_no_init.assets.create(metadata['attributes'], consumer)

    metadata = workflow_ddo['service'][0]
    metadata['attributes']['main']['workflow']['stages'][0]['input'][0]['id'] = ddo_computing.did
    metadata['attributes']['main']['workflow']['stages'][0]['transformation']['id'] = ddo_algorithm.did
    workflow_ddo = consumer_instance_no_init.assets.create(metadata['attributes'], publisher)
    assert workflow_ddo

    # order compute asset
    service = ddo_computing.get_service(service_type=ServiceTypes.CLOUD_COMPUTE)
    sa = ServiceAgreement.from_service_dict(service.as_dictionary())
    agreement_id = consumer_instance_no_init.assets.order(ddo_computing.did, sa.index, consumer, consumer)

    keeper = Keeper.get_instance()
    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id, 60, None, (), wait=True
    )
    assert event is not None, "Reward condition is not found"

    # execute workflow
    execution_id = consumer_instance_no_init.assets.execute(agreement_id, ddo_computing.did, sa.index, consumer,
        workflow_ddo.did)
    assert execution_id

    publisher_instance_no_init.assets.retire(ddo_computing.did, publisher)
    publisher_instance_no_init.assets.retire(ddo_algorithm.did, consumer)
    publisher_instance_no_init.assets.retire(workflow_ddo.did, publisher)


@pytest.mark.skip(reason='Compute stack disabled for now. See https://github.com/nevermined-io/internal/issues/218')
def test_compute_status(publisher_instance_no_init, consumer_instance_no_init, metadata, algorithm_ddo, workflow_ddo):
    consumer = publisher_instance_no_init.main_account
    publisher = consumer_instance_no_init.main_account

    # publish compute
    ddo_computing = publisher_instance_no_init.assets.create_compute(metadata, publisher)

    # publish algorithm
    metadata = algorithm_ddo['service'][0]
    ddo_algorithm = consumer_instance_no_init.assets.create(metadata['attributes'], consumer)

    metadata = workflow_ddo['service'][0]
    metadata['attributes']['main']['workflow']['stages'][0]['input'][0]['id'] = ddo_computing.did
    metadata['attributes']['main']['workflow']['stages'][0]['transformation']['id'] = ddo_algorithm.did
    workflow_ddo = consumer_instance_no_init.assets.create(metadata['attributes'], publisher)
    assert workflow_ddo

    # order compute asset
    service = ddo_computing.get_service(service_type=ServiceTypes.CLOUD_COMPUTE)
    sa = ServiceAgreement.from_service_dict(service.as_dictionary())
    agreement_id = consumer_instance_no_init.assets.order(ddo_computing.did, sa.index, consumer, consumer)

    keeper = Keeper.get_instance()
    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id, 60, None, (), wait=True
    )
    assert event is not None, "Reward condition is not found"

    # execute workflow
    execution_id = consumer_instance_no_init.assets.execute(agreement_id, ddo_computing.did, sa.index, consumer,
        workflow_ddo.did)

    # get status
    status = consumer_instance_no_init.assets.compute_status(agreement_id, execution_id, consumer)
    assert status

    publisher_instance_no_init.assets.retire(ddo_computing.did, publisher)
    publisher_instance_no_init.assets.retire(ddo_algorithm.did, consumer)
    publisher_instance_no_init.assets.retire(workflow_ddo.did, publisher)

@pytest.mark.skip(reason="Compute containers need an update")
def test_compute_logs(publisher_instance_no_init, consumer_instance_no_init, metadata, algorithm_ddo, workflow_ddo):
    consumer = publisher_instance_no_init.main_account
    publisher = consumer_instance_no_init.main_account

    # publish compute
    ddo_computing = publisher_instance_no_init.assets.create_compute(metadata, publisher)

    # publish algorithm
    metadata = algorithm_ddo['service'][0]
    ddo_algorithm = consumer_instance_no_init.assets.create(metadata['attributes'], consumer)

    metadata = workflow_ddo['service'][0]
    metadata['attributes']['main']['workflow']['stages'][0]['input'][0]['id'] = ddo_computing.did
    metadata['attributes']['main']['workflow']['stages'][0]['transformation']['id'] = ddo_algorithm.did
    workflow_ddo = consumer_instance_no_init.assets.create(metadata['attributes'], publisher)
    assert workflow_ddo

    # order compute asset
    service = ddo_computing.get_service(service_type=ServiceTypes.CLOUD_COMPUTE)
    sa = ServiceAgreement.from_service_dict(service.as_dictionary())
    agreement_id = consumer_instance_no_init.assets.order(ddo_computing.did, sa.index, consumer, consumer)

    keeper = Keeper.get_instance()
    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id, 60, None, (), wait=True
    )
    assert event is not None, "Reward condition is not found"

    # execute workflow
    execution_id = consumer_instance_no_init.assets.execute(agreement_id, ddo_computing.did, sa.index, consumer,
        workflow_ddo.did)

    # get logs
    logs = consumer_instance_no_init.assets.compute_logs(agreement_id, execution_id, consumer)
    assert logs

    publisher_instance_no_init.assets.retire(ddo_computing.did, publisher)
    publisher_instance_no_init.assets.retire(ddo_algorithm.did, consumer)
    publisher_instance_no_init.assets.retire(workflow_ddo.did, publisher)


def test_agreement_direct(publisher_instance, consumer_instance, metadata):
    publisher_account = publisher_instance.main_account
    consumer_account = consumer_instance.main_account
    ddo = publisher_instance.assets.create(metadata, publisher_account,
                                                   providers=[
                                                       '0x068Ed00cF0441e4829D9784fCBe7b9e26D4BD8d0'])

    agreement_id = consumer_instance.assets.order(ddo.did,
                                                                 ServiceTypesIndices.DEFAULT_ACCESS_INDEX,
                                                                 consumer_account,
                                                                 consumer_account
                                                                 )
    assert publisher_instance.agreements.status(agreement_id)

    keeper = publisher_instance.keeper

    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for LockRewardCondition.Fulfilled'

    assert publisher_instance.agreements.is_access_granted(agreement_id, ddo.did,
                                                                   consumer_account.address)
    publisher_instance.assets.retire(ddo.did, publisher_account)


def test_nfts(publisher_instance, metadata):
    publisher = publisher_instance.main_account
    someone_address = "0x00a329c0648769A73afAc7F9381E08FB43dBEA72"
    ddo = publisher_instance.assets.create(metadata, publisher, cap=100, royalties=0)

    balance = publisher_instance.nfts.balance(publisher.address, ddo.did)
    assert balance == 0
    balance_consumer = publisher_instance.nfts.balance(someone_address, ddo.did)
    assert balance_consumer == 0

    publisher_instance.nfts.mint(ddo.did, 10, account=publisher)
    assert balance + 10 == publisher_instance.nfts.balance(publisher.address, ddo.did)
    assert publisher_instance.nfts.transfer_nft(ddo.did, someone_address, 1, publisher)
    assert publisher_instance.nfts.balance(publisher.address, ddo.did) == 9
    assert publisher_instance.nfts.balance(someone_address, ddo.did) == balance_consumer + 1
    publisher_instance.nfts.burn(ddo.did, 9, account=publisher)
    assert balance == publisher_instance.nfts.balance(publisher.address, ddo.did)
    publisher_instance.assets.retire(ddo.did, publisher)


import json
import logging
import os

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes
from common_utils_py.did import did_to_id

logger = logging.getLogger(__name__)


class AssetConsumer:

    @staticmethod
    def access(service_agreement_id, service_index, ddo, consumer_account, destination,
                 gateway, secret_store, config, index=None, service_type=ServiceTypes.ASSET_ACCESS):
        """
        Download asset data files or result files from a compute job.

        :param service_agreement_id: Service agreement id, str
        :param service_index: identifier of the service inside the asset DDO, str
        :param ddo: DDO
        :param consumer_account: Account instance of the consumer
        :param destination: Path, str
        :param gateway: Gateway instance
        :param secret_store: SecretStore instance
        :param config: Sdk configuration instance
        :param index: Index of the document that is going to be downloaded, int
        :param service_type: service type
        :return: Asset folder path, str
        """
        did = ddo.did
        sa = ServiceAgreement.from_ddo(service_type, ddo)
        consume_url = sa.service_endpoint
        if not consume_url:
            logger.error(
                'Consume asset failed, service definition is missing the "serviceEndpoint".')
            raise AssertionError(
                'Consume asset failed, service definition is missing the "serviceEndpoint".')

        if ddo.get_service(ServiceTypes.AUTHORIZATION):
            secret_store_service = ddo.get_service(service_type=ServiceTypes.AUTHORIZATION)
            secret_store_url = secret_store_service.service_endpoint
            secret_store.set_secret_store_url(secret_store_url)

        asset_folder = create_asset_folder(did, service_index, destination)

        if index is not None:
            assert isinstance(index, int), logger.error('index has to be an integer.')
            assert index >= 0, logger.error('index has to be 0 or a positive integer.')
            assert index < len(ddo.metadata['main']['files']), logger.error(
                'index can not be bigger than the number of files')

        if service_type is ServiceTypes.NFT_ACCESS:
            uri = '/nft-access'
        else:
            uri = '/access'

        if index is not None:
            gateway.access_service(
                did,
                service_agreement_id,
                consume_url,
                consumer_account,
                asset_folder,
                config,
                index,
                uri
            )
        else:
            for i, _file in enumerate(ddo.metadata['main']['files']):
                gateway.access_service(
                    did,
                    service_agreement_id,
                    consume_url,
                    consumer_account,
                    asset_folder,
                    config,
                    i,
                    uri
                )

        return asset_folder

    @staticmethod
    def download(service_index, ddo, owner_account, destination,
                 gateway, secret_store, config, index=None):
        """
        Allows the onwer of and asset to download the data files.

        :param service_index: identifier of the service inside the asset DDO, str
        :param ddo: DDO
        :param owner_account: Account instance of the owner
        :param destination: Path, str
        :param gateway: Gateway instance
        :param secret_store: SecretStore instance
        :param config: Sdk configuration instance
        :param index: Index of the document that is going to be downloaded, int
        :return: Asset folder path, str
        """
        did = ddo.did
        if ddo.get_service(ServiceTypes.AUTHORIZATION):
            secret_store_service = ddo.get_service(service_type=ServiceTypes.AUTHORIZATION)
            secret_store_url = secret_store_service.service_endpoint
            secret_store.set_secret_store_url(secret_store_url)

        asset_folder = create_asset_folder(did, service_index, destination)

        if index is not None:
            assert isinstance(index, int), logger.error('index has to be an integer.')
            assert index >= 0, logger.error('index has to be 0 or a positive integer.')
            assert index < len(ddo.metadata['main']['files']), logger.error(
                'index can not be bigger than the number of files')
        if index is not None:
            gateway.download(
                did,
                owner_account,
                asset_folder,
                index,
                config
            )
        else:
            for i, _file in enumerate(ddo.metadata['main']['files']):
                gateway.download(
                    did,
                    owner_account,
                    asset_folder,
                    i,
                    config
                )

        return asset_folder

    @staticmethod
    def compute_logs(service_agreement_id, execution_id, account, gateway, config):
        """
        Get the logs of a compute workflow.

        :param service_agreement_id: The id of the service agreement that ordered the compute job, str
        :param execution_id: The id of the compute job, str
        :param account: Account instance that ordered the execution of the compute job
        :param gateway: Gateway instance
        :param config: Sdk configuration instance
        :return: list, compute logs
        """
        return gateway.compute_logs(service_agreement_id, execution_id, account, config).json()

    @staticmethod
    def compute_status(service_agreement_id, execution_id, account, gateway, config):
        """
        Get the status of a compute workflow.

        :param service_agreement_id: The id of the service agreement that ordered the compute job, str
        :param execution_id: The id of the compute job, str
        :param account: Account instance that ordered the execution of the compute job
        :param gateway: Gateway instance
        :param config: Sdk configuration instance
        :return: dict, compute logs
        """
        return gateway.compute_status(service_agreement_id, execution_id, account, config).json()

def create_asset_folder(did, service_index, destination):
    """
    Creates the asset folder to download the assets

    :param did: the identifier of the asset ddo.
    :param service_index: identifier of the service inside the asset DDO, str
    :param destination: Path, str
    :return: Asset folder path, str
    """
    if not os.path.isabs(destination):
        destination = os.path.abspath(destination)
    if not os.path.exists(destination):
        os.mkdir(destination)

    asset_folder = os.path.join(destination,
                                f'datafile.{did_to_id(did)}.{service_index}')
    if not os.path.exists(asset_folder):
        os.mkdir(asset_folder)

    return asset_folder



import logging
from nevermined_sdk_py.nevermined.files import Files

from common_utils_py.did_resolver.did_resolver import DIDResolver
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.web3_provider import Web3Provider

from nevermined_sdk_py.assets.asset_consumer import AssetConsumer
from nevermined_sdk_py.assets.asset_executor import AssetExecutor
from nevermined_sdk_py.config_provider import ConfigProvider
from nevermined_sdk_py.faucet.faucet import Faucet
from nevermined_sdk_py.nevermined.accounts import Accounts
from nevermined_sdk_py.nevermined.agreements import Agreements
from nevermined_sdk_py.nevermined.assets import Assets
from nevermined_sdk_py.nevermined.auth import Auth
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from nevermined_sdk_py.nevermined.nfts import NFTs
from nevermined_sdk_py.nevermined.provenance import Provenance
from nevermined_sdk_py.nevermined.providers import Providers
from nevermined_sdk_py.nevermined.secret_store import SecretStore
from nevermined_sdk_py.nevermined.services import Services
from nevermined_sdk_py.nevermined.templates import Templates
from nevermined_sdk_py.nevermined.tokens import Tokens

CONFIG_FILE_ENVIRONMENT_NAME = 'CONFIG_FILE'

logger = logging.getLogger(__name__)


class Nevermined:
    """The Nevermined class is the entry point into Nevermined Protocol."""

    def __init__(self, config=None):
        """
        Initialize Nevermined class.
           >> # Make a new Nevermined instance
           >> nevermined = Nevermined({...})

        This class provides the main top-level functions in nevermined protocol:
         * Publish assets metadata and associated services
            * Each asset is assigned a unique DID and a DID Document (DDO)
            * The DDO contains the asset's services including the metadata
            * The DID is registered on-chain with a URL of the metadata store
              to retrieve the DDO from

            >> ddo = nevermined.assets.create(metadata, publisher_account)

         * Discover/Search assets via the current configured metadata store.
            >> assets_list = nevermined.assets.search('search text')

         * Purchase asset services by choosing a service agreement from the
           asset's DDO. Purchase goes through the service agreements interface
           and starts by signing a service agreement then sending the signature
           to the publisher's Gateway server via the `purchaseEndpoint` in the service
           definition:

           >> service_def_id = ddo.get_service(ServiceTypes.ASSET_ACCESS).service_definition_id
           >> service_agreement_id = nevermined.assets.order(did, service_def_id, consumer_account)

        An instance of Nevermined is parameterized by a `Config` instance.

        :param config: Config instance
        """
        # Configuration information for the market is stored in the Config class
        # config = Config(filename=config_file, options_dict=config_dict)
        if not config:
            config = ConfigProvider.get_config()

        self._config = config
        self._web3 = Web3Provider.get_web3(self._config.keeper_url)
        ContractHandler.artifacts_path = self._config.keeper_path
        contracts = [
            'DIDRegistry',
            'Dispenser',
            'TemplateStoreManager',
            'NeverminedToken',
            'ConditionStoreManager',
            'EscrowAccessSecretStoreTemplate',
            'AgreementStoreManager',
            'AgreementStoreManager',
            'AccessSecretStoreCondition',
            'LockRewardCondition',
            'HashLockCondition',
            'SignCondition',
            'EscrowReward'
        ]
        self._keeper = Keeper.get_instance(self._config.keeper_path, contracts)
        self._did_resolver = DIDResolver(self._keeper.did_registry)

        # Initialize the public sub-modules
        self.tokens = Tokens(self._keeper)
        self.accounts = Accounts(self._keeper, self._config, self.tokens, Faucet)
        self.templates = Templates(
            self._keeper,
            config
        )
        self.agreements = self._make_agreements()
        self.assets = Assets(
            self._keeper,
            self._did_resolver,
            self.agreements,
            AssetConsumer,
            AssetExecutor,
            self._config
        )
        self.nfts = NFTs(self._keeper)
        self.services = Services()
        self.secret_store = SecretStore(self._config)
        self.providers = Providers(self._keeper, self._did_resolver, self._config)
        self.auth = Auth(self._keeper, self._config.storage_path)
        self.provenance = Provenance(self._keeper, self._config)
        self.files = Files(self._config)

        logger.debug('Nevermined instance initialized: ')
        logger.debug(f'\tOther accounts: {sorted([a.address for a in self.accounts.list()])}')
        logger.debug(f'\tDIDRegistry @ {self._keeper.did_registry.address}')

    @property
    def config(self):
        return self._config

    @property
    def keeper(self):
        """Keeper instance."""
        return self._keeper

    def _make_agreements(self):
        return Agreements(
            self._keeper,
            self._did_resolver,
            AssetConsumer,
            AssetExecutor,
            self._config
        )



import tempfile
import pathlib
import uuid
import pytest

from tests.resources.helper_functions import get_publisher_instance

pytestmark = pytest.mark.skip(reason="Requires Powergate in the CI")

def test_upload_filecoin():
    publisher_instance = get_publisher_instance(True, False, False)

    with tempfile.NamedTemporaryFile() as f:
        f.write(b"Hello, Nevermined!")
        f.flush()
        url = publisher_instance.files.upload_filecoin(f.name)

    assert url == "cid://QmSJA3xNH62sj4xggZZzCp2VXpsXbkR9zYoqNYXp3c4xuN"


def test_create_asset_with_cid(metadata):
    publisher_instance = get_publisher_instance(True, False, False)
    publisher = publisher_instance.main_account

    with tempfile.NamedTemporaryFile() as f:
        f.write(b"aaa")
        f.flush()
        url = publisher_instance.files.upload_filecoin(f.name)

    metadata["main"]["files"] = [
        {
            "index": 0,
            "contentType": "text/text",
            "encoding": "UTF-8",
            "compression": "zip",
            "checksum": str(uuid.uuid4()),
            "checksumType": "MD5",
            "contentLength": "12057507",
            "url": url,
        }
    ]
    ddo = publisher_instance.assets.create(metadata, publisher)
    assert ddo is not None


def test_download_asset_with_cid(metadata):
    publisher_instance = get_publisher_instance(True, False, False)
    publisher = publisher_instance.main_account

    with tempfile.NamedTemporaryFile() as f:
        f.write(b"bbb")
        f.flush()
        url = publisher_instance.files.upload_filecoin(f.name)

    metadata["main"]["files"] = [
        {
            "index": 0,
            "contentType": "text/text",
            "encoding": "UTF-8",
            "compression": "zip",
            "checksum": str(uuid.uuid4()),
            "checksumType": "MD5",
            "contentLength": "12057507",
            "url": url,
        }
    ]
    ddo = publisher_instance.assets.create(metadata, publisher)

    service_agreement = ddo.get_service("access")
    path = publisher_instance.assets.download(
        ddo.did, service_agreement.index, publisher, "/tmp/outputs"
    )
    assert path is not None

    with (pathlib.Path(path) / "file-0").open("rb") as f:
        assert f.read() == b"bbb"



import time

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes
from common_utils_py.ddo.ddo import DDO

from examples import ExampleConfig
from nevermined_sdk_py import ConfigProvider
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from tests.resources.helper_functions import (get_consumer_account, get_publisher_account,
                                              log_event, get_registered_ddo_nft)


def test_nfts_flow(publisher_instance_no_init, consumer_instance_no_init):
    config = ExampleConfig.get_config()
    keeper = Keeper.get_instance()

    pub_acc = get_publisher_account()

    # Register ddo
    ddo = get_registered_ddo_nft(publisher_instance_no_init, pub_acc)
    asset_id = ddo.asset_id
    nft_amounts = 1
    assert isinstance(ddo, DDO)

    consumer_account = get_consumer_account()

    consumer_instance_no_init.accounts.request_tokens(consumer_account, 100)

    service_sales = ddo.get_service(service_type=ServiceTypes.NFT_SALES)
    sa_sales = ServiceAgreement.from_service_dict(service_sales.as_dictionary())

    amounts = sa_sales.get_amounts_int()
    receivers = sa_sales.get_receivers()
    number_nfts = sa_sales.get_number_nfts()
    nft_contract_address = sa_sales.get_nft_contract_address()
    nft_transfer = sa_sales.get_nft_transfer_or_mint()
    token_address = keeper.token.address

    sales_agreement_id = consumer_instance_no_init.assets.order(
        ddo.did, sa_sales.index, consumer_account, consumer_account)

    sales_agreement = keeper.agreement_manager.get_agreement(sales_agreement_id)
    assert sales_agreement.did == asset_id, ''

    lock_cond_id = sales_agreement.condition_ids[0]
    access_cond_id = sales_agreement.condition_ids[1]
    escrow_cond_id = sales_agreement.condition_ids[2]

    # transfer the nft
    keeper.transfer_nft_condition.fulfill(
        sales_agreement_id,
        asset_id,
        consumer_account.address,
        nft_amounts,
        lock_cond_id,
        nft_contract_address,
        nft_transfer,
        pub_acc
    )

    # escrow payment
    keeper.escrow_payment_condition.fulfill(
        sales_agreement_id,
        asset_id,
        amounts,
        receivers,
        consumer_account.address,
        keeper.escrow_payment_condition.address,
        token_address,
        lock_cond_id,
        access_cond_id,
        pub_acc
    )

    assert keeper.condition_manager.get_condition_state(lock_cond_id) == 2, ''
    assert keeper.condition_manager.get_condition_state(access_cond_id) == 2, ''
    assert keeper.condition_manager.get_condition_state(escrow_cond_id) == 2, ''

    assert keeper.nft_upgradeable.balance(consumer_account.address, asset_id) >= number_nfts


    # CHECK ACCESS CREATING THE AGREEMENT
    nft_access_service_agreement = ServiceAgreement.from_ddo(ServiceTypes.NFT_ACCESS, ddo)
    sa_access = ddo.get_service(service_type=ServiceTypes.NFT_ACCESS)
    nft_access_agreement_id = consumer_instance_no_init.assets.order(
        ddo.did, sa_access.index, consumer_account, consumer_account)

    event = keeper.nft_access_template.subscribe_agreement_created(
        nft_access_agreement_id,
        10,
        log_event(keeper.nft_access_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '

    time.sleep(3)
    keeper.nft_holder_condition.fulfill(
        nft_access_agreement_id, asset_id, consumer_account.address, number_nfts, consumer_account
    )
    time.sleep(3)
    keeper.nft_access_condition.fulfill(
        nft_access_agreement_id, asset_id, consumer_account.address, pub_acc
    )
    time.sleep(3)

    access_agreement = keeper.agreement_manager.get_agreement(nft_access_agreement_id)
    assert access_agreement.did == asset_id, ''

    nft_holder_cond_id = access_agreement.condition_ids[0]
    nft_access_cond_id = access_agreement.condition_ids[1]

    assert keeper.condition_manager.get_condition_state(nft_access_cond_id) == 2, ''
    assert keeper.condition_manager.get_condition_state(nft_holder_cond_id) == 2, ''

    # AND HERE TESTING WITHOUT CREATING THE AGREEMENT
    service_access = ddo.get_service(service_type=ServiceTypes.NFT_ACCESS)

    no_agreement_id = '0x'


    assert consumer_instance_no_init.assets.access(
        no_agreement_id,
        ddo.did,
        service_access.index,
        consumer_account,
        config.downloads_path,
        service_type=ServiceTypes.NFT_ACCESS
    )

    # AND HERE CHECKING CREATING AN AGREEMENT FIRST
    nft_access_service_agreement = ServiceAgreement.from_ddo(ServiceTypes.NFT_ACCESS, ddo)
    nft_access_agreement_id_seed = ServiceAgreement.create_new_agreement_id()

    ((nft_access_agreement_id_seed, nft_access_agreement_id), *conditions) = nft_access_service_agreement.generate_agreement_condition_ids(
        nft_access_agreement_id_seed, asset_id, consumer_account.address, keeper, pub_acc.address)

    (nft_access_cond_id, nft_holder_cond_id) = conditions

    result = keeper.nft_access_template.create_agreement(
        nft_access_agreement_id_seed,
        asset_id,
        [nft_holder_cond_id[0], nft_access_cond_id[0]],
        nft_access_service_agreement.conditions_timelocks,
        nft_access_service_agreement.conditions_timeouts,
        consumer_account.address,
        pub_acc
    )
    assert result is True

    event = keeper.nft_access_template.subscribe_agreement_created(
        nft_access_agreement_id,
        10,
        log_event(keeper.nft_access_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '

    # This is because in this test the gateway is executed using the same account than the consumer
    keeper.did_registry.add_provider(asset_id, consumer_account.address, pub_acc)
    time.sleep(3)

    assert consumer_instance_no_init.assets.access(
        nft_access_agreement_id,
        ddo.did,
        service_access.index,
        consumer_account,
        config.downloads_path,
        index=0,
        service_type=ServiceTypes.NFT_ACCESS
    )
